# Imports.

import numpy as np
import pandas as pd
from scipy.stats import pearsonr
from sklearn.preprocessing import StandardScaler
import matplotlib.pyplot as plt
from scipy.optimize import curve_fit

from . import Preprocessing
from ..Global_Analysis import Utilities as util2
from .. import Global_Utilities as gu

# Function definitions.

def compute_derivatives( data, width = 1 ):
    '''Compute the derivatives.'''

    derivative_data = [data[0][width: -width], [], []]

    for i in range( len( data[1] ) ):

        derivative_data[1].append( np.array( gu.derivative( data[0], data[1][i], width ) ) )

    for i in range( len( data[2] ) ):

        derivative_data[2].append( np.array( gu.derivative( data[0], data[2][i], width ) ) )

    return derivative_data

def identify_peaks( data, first_derivative_data, second_derivative_data, third_derivative_data, peak_threshold_1, peak_threshold_2 ):
    '''Identify peaks in the spectrums by considering where the curve of the third derivative
    crosses the x-axis with positive gradient as wavenumbers increase,
    using the magnitude of the second derivative at this point to determine the significance.'''

    y_peaks, m_peaks, mag_y_peaks, mag_m_peaks = [], [], [], []

    structure = [(third_derivative_data[0], data[1], first_derivative_data[1], second_derivative_data[1], third_derivative_data[1], y_peaks, mag_y_peaks), (third_derivative_data[0], data[2], first_derivative_data[2], second_derivative_data[2], third_derivative_data[2], m_peaks, mag_m_peaks)]

    for s in structure:

        xd3, d0, d1, d2, d3, _, _ = s

        for i in range( len( d3 ) ):

            local_min, local_mag = [], []

            for j in range( 1, len( d3[i] ) ):

                if d3[i][j] <= 0 and d3[i][j - 1] > 0:

                    if d2[i][j + 1] < peak_threshold_1 or d2[i][j] < peak_threshold_1:

                        if d2[i][j + 1] < d2[i][j]:

                            if not (d2[i][j] > 0 or d2[i][j + 2] > 0): # Ensure peak not due to noise/more precise FTIR spectra.

                                local_min.append( xd3[j] )
                                local_mag.append( d2[i][j + 1] )

                        else:

                            if not (d2[i][j - 1] > 0 or d2[i][j + 1] > 0): # Ensure peak not due to noise/more precise FTIR spectra.

                                local_min.append( xd3[j - 1] )
                                local_mag.append( d2[i][j] )

                # The following clause was used to find minima in the spectra. Was particularly useful for CO2 where there were minima, I think due to background spectra being subtracted. There was a strong correlation
                # between 667, 2343, and 2381.

                # elif d3[i][j] >= 0 and d3[i][j - 1] < 0:
                #
                #     if d2[i][j + 1] > -peak_threshold_1 or d2[i][j] > -peak_threshold_1:
                #
                #         if d2[i][j + 1] > d2[i][j]:
                #
                #             local_min.append( xd3[j] )
                #             local_mag.append( -d2[i][j + 1] )
                #
                #         else:
                #
                #             local_min.append( xd3[j - 1] )
                #             local_mag.append( -d2[i][j] )

            s[5].append( local_min )
            s[6].append( local_mag )

    return y_peaks, m_peaks, mag_y_peaks, mag_m_peaks

def linear_filter( magnitude, cutoff, threshold ):
    '''Assigning a sigmoid filter to the magnitude of peaks.'''

    for i in range( len( magnitude ) ):

        for j in range( len( magnitude[i] ) ):

            if magnitude[i][j] < -cutoff:

                magnitude[i][j] = -1

            else:

                magnitude[i][j] = -(-magnitude[i][j] + threshold) / (cutoff + threshold)

def merge_close_peaks( peaks_array, peak_spacing ):
    '''Peaks whose wavenumbers are within a threshold are merged.'''

    iterations = 2

    spacings = [3, peak_spacing] # Was [3, peak_spacing].

    unique, counts = np.unique( peaks_array, return_counts = True )

    dict_peaks = dict( zip( unique, counts ) )

    popped_dict = {}

    for r in range( iterations ):

        skip = 0

        for ind, u in enumerate( unique ):

            if skip > 0:

                skip -= 1
                continue

            if ind >= len( unique ) - 1:

                continue

            if unique[ind + 1] - unique[ind] > spacings[r]:

                continue

            else:

                indices = [ind, ind + 1]

                for i in range( ind + 2, len( unique ) ):

                    if unique[i] - unique[i - 1] < spacings[r]:

                        indices.append( i )

                    else:

                        break

                skip = len( indices ) - 1
                index_of_max_count = indices[0]

                for i in range( 1, len( indices ) ):

                    if counts[indices[i]] > counts[index_of_max_count]:

                        index_of_max_count = indices[i]

                for i in indices:

                    if i == index_of_max_count:

                        continue

                    else:

                        dict_peaks[unique[index_of_max_count]] += counts[i]
                        dict_peaks.pop( unique[i] )
                        popped_dict[unique[i]] = unique[index_of_max_count]

        unique = np.array( list( dict_peaks.keys() ) )
        counts = np.array( list( dict_peaks.values() ) )

    return dict_peaks, popped_dict

def populate_peaks_array( list_of_peaks, magnitude_of_peaks, peak_spacing, peak_limit ):
    '''Produce a peaks array: where each row is a heat map of where peaks are for a corresponding specturm.'''

    peaks = []

    for i in range( len( list_of_peaks ) ):

        for j in list_of_peaks[i]:

            peaks.append( j )

    peaks_array = np.array( peaks )

    dict_peaks, popped_dict = merge_close_peaks( peaks_array, peak_spacing )

    peaks = list( dict_peaks.keys() )

    peaks_array = np.zeros( (len( list_of_peaks ), len( peaks )) )

    for i in range( len( list_of_peaks ) ):

        for ind, j in enumerate( list_of_peaks[i] ):

            if j in peaks:

                if -magnitude_of_peaks[i][ind] > peaks_array[i][peaks.index( j )]:

                    peaks_array[i][peaks.index( j )] = -magnitude_of_peaks[i][ind]

            elif popped_dict[j] in peaks:

                if -magnitude_of_peaks[i][ind] > peaks_array[i][peaks.index( popped_dict[j] )]:

                    peaks_array[i][peaks.index( popped_dict[j] )] = -magnitude_of_peaks[i][ind]

            else:

                if -magnitude_of_peaks[i][ind] > peaks_array[i][peaks.index( popped_dict[popped_dict[j]] )]:

                    peaks_array[i][peaks.index( popped_dict[popped_dict[j]] )] = -magnitude_of_peaks[i][ind]

    peaks = list( np.array( peaks )[np.where( np.array( peaks ) < peak_limit )[0]] )
    peaks_array = peaks_array[:, :len( peaks )]

    return peaks, peaks_array

def remove_identical_columns( peaks_array, peaks ):
    '''Remove identical columns in a peaks array.'''

    trans_arr = peaks_array.T
    not_uniform_column = []

    for i in range( trans_arr.shape[0] ):

        if not np.all( trans_arr[i] == trans_arr[i][0] ):

            not_uniform_column.append( i )

    peaks = list( np.array( peaks )[not_uniform_column] )
    peaks_array = peaks_array[:, not_uniform_column]

    return peaks, peaks_array

def remove_wavenumbers( peaks, peaks_array ):

    # Remove wavenumbers corresponding to CO2.

    peaks_to_remove = []

    for i, p in enumerate( peaks ):

        if p < 675 and p > 660:

            peaks_to_remove.append( i )

        elif p < 2400 and p > 2300:

            peaks_to_remove.append( i )

    peaks_to_remove.reverse()

    for p in peaks_to_remove:

        peaks.pop( p )
        peaks_array = np.delete( peaks_array, p, 1 )

    return peaks_array

def extract_FTIR_features( output_directory, file_data, data, first_derivative_data, second_derivative_data, third_derivative_data, peak_threshold_1, peak_threshold_2, peak_spacing, peak_limit ):

    m_peak_analysis = True
    y_peak_analysis = True

    y_peaks, m_peaks, mag_y_peaks, mag_m_peaks = identify_peaks( data, first_derivative_data, second_derivative_data, third_derivative_data, peak_threshold_1, peak_threshold_2 )

    m_peaks_array, y_peaks_array = [], []

    if m_peak_analysis:

        linear_filter( mag_m_peaks, peak_threshold_2, peak_threshold_1 )
        m_peaks, m_peaks_array = populate_peaks_array( m_peaks, mag_m_peaks, peak_spacing, peak_limit )
        m_peaks, m_peaks_array = remove_identical_columns( m_peaks_array, m_peaks )
        m_peaks_array = remove_wavenumbers( m_peaks, m_peaks_array )

        feature_names = ["FTIR_{}".format( int( i ) ) for i in m_peaks]

        df = gu.array_with_column_titles_to_df( m_peaks_array, feature_names )

        df.to_csv( output_directory + "FTIR/Features/Features.csv" )

    if y_peak_analysis:

        peaks_specified_by_m_peak_analysis = True

        linear_filter( mag_y_peaks, peak_threshold_2, peak_threshold_1 )

        if peaks_specified_by_m_peak_analysis:

            peaks = m_peaks # Current m_peaks are [777.4173764906302, 808.2725724020441, 825.6286201022145, 841.0562180579215, 854.5553662691651, 875.7683134582622, 908.5519591141395, 972.190800681431, 997.2606473594548, 1018.473594548552, 1066.6848381601362, 1082.1124361158431, 1166.9642248722316, 1195.8909710391822, 1213.2470187393524, 1261.4582623509368, 1367.5229982964224, 1377.1652470187391, 1398.3781942078365, 1409.9488926746167, 1435.0187393526403, 1490.9437819420782, 1541.083475298126, 1577.72402044293, 1631.7206132879044, 1645.2197614991483, 1658.7189097103917, 1739.7137989778535, 1978.841567291312, 2162.044293015332, 2952.708688245315, 2960.4224872231684, 3297.9011925042582].

            peaks_array = np.zeros( (len( y_peaks ), len( peaks )) )

            for i in range( len( y_peaks ) ):

                for j in range( len( peaks ) ):

                    argmin = min( range( len( y_peaks[i] ) ), key = lambda x: abs( y_peaks[i][x] - peaks[j] ) )

                    if abs( y_peaks[i][argmin] - peaks[j] ) < 6:

                        peaks_array[i][peaks.index( peaks[j] )] = -mag_y_peaks[i][argmin]

            y_peaks = peaks
            y_peaks_array = peaks_array

        else:

            y_peaks, y_peaks_array = populate_peaks_array( y_peaks, mag_y_peaks, peak_spacing, peak_limit )
            y_peaks, y_peaks_array = remove_identical_columns( y_peaks_array, y_peaks )
            y_peaks_array = remove_wavenumbers( y_peaks, y_peaks_array )

        feature_names = ["FTIR_{}".format( int( i ) ) for i in y_peaks]

        df = gu.array_with_column_titles_to_df( y_peaks_array, feature_names )

        df.to_csv( output_directory + "FTIR/Features/Specimen_Features.csv" )

    return m_peaks, m_peaks_array, y_peaks, y_peaks_array

def read_and_analyse_FTIR_features( ip, file_data ):

    precomputed_means = False

    plot_specimen_bars = False
    plot_mean_bars = False
    plot_specimen_features = True
    plot_mean_features = True
    plot_specimen_distance_matrix = True
    plot_mean_distance_matrix = True
    plot_specimen_dendrogram = True
    plot_mean_dendrogram = True

    resin_data = gu.get_list_of_resins_data( ip.directory )

    feature_names, features = gu.csv_to_df_to_array_and_column_titles( ip.output_directory + "FTIR/Features/Specimen_Features.csv" )

    sample, sample_array, samples_present, samples_present_array = gu.sample_data_from_file_data( file_data )

    if not ip.sample_mask:

        # ip.sample_mask = [11, 14, 10, 4, 13, 21, 23, 18, 22, 20, 2, 3, 17, 16, 19, 1, 15, 12, 6, 5, 7, 9, 8, 24]
        # ip.sample_mask = [301, 303, 306, 308, 309, 312, 313, 315, 318, 320, 324, 325]
        ip.sample_mask = [11, 14, 10, 4, 13, 21, 23, 18, 22, 20, 2, 3, 17, 16, 19, 1, 15, 12, 6, 5, 7, 9, 8, 24, 500, 502, 504, 506, 508, 510, 512, 514, 516, 518]

    sample_mask = ip.sample_mask

    sample_mask = gu.remove_redundant_samples( sample_mask, samples_present )

    if not ip.feature_selection:

        ip.feature_selection = [i for i in range( len( feature_names ) )]

    feature_selection = ip.feature_selection

    features = features[:, feature_selection]
    feature_names = list( np.array( feature_names )[feature_selection] )

    mean_features_unnormalised = gu.extract_mean_features( features, sample_array, samples_present )
    mean_feature_names = feature_names.copy()

    mean_features_unnormalised_plus_sample_mask = np.hstack( (samples_present_array[:, np.newaxis], mean_features_unnormalised) )

    df = gu.array_with_column_titles_to_df( mean_features_unnormalised_plus_sample_mask, ["sample"] + mean_feature_names )

    df.to_csv( ip.output_directory + "FTIR/Features/Mean_Features_Unnormalised.csv" )

    std_of_features = gu.extract_std_of_features( features, sample_array, samples_present )

    std_of_features_plus_sample_mask = np.hstack( (samples_present_array[:, np.newaxis], std_of_features) )

    df = gu.array_with_column_titles_to_df( std_of_features_plus_sample_mask, ["sample"] + mean_feature_names )

    df.to_csv( ip.output_directory + "FTIR/Features/Std_of_Features_Unnormalised.csv" )

    specimen_mask = gu.produce_mask( sample_array, sample_mask )

    features = features[specimen_mask, :]
    file_data_mask = np.array( file_data )[specimen_mask] # Orders file data according to specimen mask.
    file_data_mask = [[int( f[0] ), int( f[1] ), f[2], f[3]] for f in file_data_mask] # Converts file data so that f[0], etc. are integers and not np.str.
    sample_array = sample_array[specimen_mask]

    feature_weights = [1 for i in range( len( feature_selection ) )]

    # gu.normalise_features( features, feature_weights )

    mean_features = gu.extract_mean_features( features, sample_array, sample_mask )
    mean_feature_names = feature_names.copy()

    std_of_features = gu.extract_std_of_features( features, sample_array, sample_mask )

    distance_matrix = gu.distance_matrix_from_features( features )

    mean_distance_matrix = gu.distance_matrix_from_features( mean_features )

    mean_features_plus_sample_mask = np.hstack( (np.array( sample_mask )[:, np.newaxis], mean_features) )

    df = gu.array_with_column_titles_to_df( mean_features_plus_sample_mask, ["sample"] + mean_feature_names )

    df.to_csv( ip.output_directory + "FTIR/Features/Mean_Features.csv" )

    std_of_features_plus_sample_mask = np.hstack( (np.array( sample_mask )[:, np.newaxis], std_of_features) )

    df = gu.array_with_column_titles_to_df( std_of_features_plus_sample_mask, ["sample"] + mean_feature_names )

    df.to_csv( ip.output_directory + "FTIR/Features/Std_of_Features.csv" )

    df = gu.array_with_column_titles_and_label_titles_to_df( mean_distance_matrix, sample_mask, sample_mask )

    df.to_csv( ip.output_directory + "FTIR/Features/Distance_Matrix.csv" )

    if plot_specimen_bars:

        for i in range( len( features[0] ) ):

            gu.plot_barchart_of_feature( features[:, i], [f[2] for f in file_data_mask], colour = True, colour_mask = sample_array, filename = ip.output_directory + "FTIR/Feature_Bars/Specimen/" + feature_names[i] + ".pdf", savefig = True )

    if plot_mean_bars:

        for i in range( len( mean_features[0] ) ):

            gu.plot_barchart_of_feature( mean_features[:, i], [resin_data.loc[i]["Label"] for i in sample_mask], errorbars = True, std = std_of_features[:, i], colour = True, colour_mask = sample_mask, filename = ip.output_directory + "FTIR/Feature_Bars/Mean/" + mean_feature_names[i] + ".pdf", savefig = True )

    if plot_specimen_features:

        gu.plot_features( ip.output_directory, features, [f[5:] for f in feature_names], [f[2] for f in file_data_mask], specimen = True, subdirectory = "FTIR/Features/", title = "Specimen_Features.pdf" )

    if plot_mean_features:

        gu.plot_features( ip.output_directory, mean_features, [f[5:] for f in mean_feature_names], [resin_data.loc[i]["Label"] for i in sample_mask], subdirectory = "FTIR/Features/", title = "Means_Features.pdf" )

    if plot_specimen_distance_matrix:

        gu.plot_distance_matrix( ip.output_directory, distance_matrix, [f[2] for f in file_data_mask], specimen = True, file_data = file_data_mask, sample_mask = sample_mask, subdirectory = "FTIR/Features/", title = "Specimen_Distance_Matrix.pdf" )

    if plot_mean_distance_matrix:

        gu.plot_distance_matrix( ip.output_directory, mean_distance_matrix, [resin_data.loc[i]["Label"] for i in sample_mask], subdirectory = "FTIR/Features/", title = "Means_Distance_Matrix.pdf" )

    if plot_specimen_dendrogram:

        gu.plot_dendrogram( ip.output_directory, distance_matrix, [f[2] for f in file_data_mask], specimen = True, subdirectory = "FTIR/Features/", title = "Specimen_Dendrogram.pdf" )

    if plot_mean_dendrogram:

        gu.plot_dendrogram( ip.output_directory, mean_distance_matrix, [resin_data.loc[i]["Label"] for i in sample_mask], subdirectory = "FTIR/Features/", title = "Means_Dendrogram.pdf" )

    if precomputed_means:

        feature_names, features = gu.csv_to_df_to_array_and_column_titles( ip.output_directory + "FTIR/Features/Features.csv" )

        feature_selection = [i for i in range( len( feature_names ) )]

        features = features[:, feature_selection]
        feature_names = list( np.array( feature_names )[feature_selection] )

        samples_present_mask = gu.produce_mask( samples_present_array, sample_mask )

        features = features[samples_present_mask, :]

        feature_weights = [1 for i in range( len( feature_selection ) )]

        # gu.normalise_features( features, feature_weights )

        distance_matrix = gu.distance_matrix_from_features( features )

        mean_features_plus_sample_mask = np.hstack( (np.array( sample_mask )[:, np.newaxis], features) )

        df = gu.array_with_column_titles_to_df( mean_features_plus_sample_mask, ["sample"] + feature_names )

        df.to_csv( ip.output_directory + "FTIR/Features/Mean_Features.csv" )

        df = gu.array_with_column_titles_and_label_titles_to_df( distance_matrix, sample_mask, sample_mask )

        df.to_csv( ip.output_directory + "FTIR/Features/Distance_Matrix.csv" )

        if plot_mean_bars:

            for i in range( len( features[0] ) ):

                gu.plot_barchart_of_feature( features[:, i], [resin_data.loc[i]["Label"] for i in sample_mask], colour = True, colour_mask = sample_mask, filename = ip.output_directory + "FTIR/Feature_Bars/Mean/" + feature_names[i] + ".pdf", savefig = True )

        if plot_mean_features:

            gu.plot_features( ip.output_directory, features, [f[5:] for f in feature_names], [resin_data.loc[i]["Label"] for i in sample_mask], subdirectory = "FTIR/Features/", title = "Means_Features.pdf" )

        if plot_mean_distance_matrix:

            gu.plot_distance_matrix( ip.output_directory, distance_matrix, [resin_data.loc[i]["Label"] for i in sample_mask], subdirectory = "FTIR/Features/", title = "Means_Distance_Matrix.pdf" )

        if plot_mean_dendrogram:

            gu.plot_dendrogram( ip.output_directory, distance_matrix, [resin_data.loc[i]["Label"] for i in sample_mask], subdirectory = "FTIR/Features/", title = "Means_Dendrogram.pdf" )

def integral_analysis( directory, output_directory, file_data, data ):
    '''Perform an analysis on the data that focuses on integrals.'''

    plot_specimen_distance_matrix = True
    plot_mean_distance_matrix = True
    plot_specimen_dendrogram = True
    plot_mean_dendrogram = True

    resin_data = gu.get_list_of_resins_data( directory )

    sample, sample_array, samples_present, samples_present_array = gu.sample_data_from_file_data( file_data )

    sample_mask = [11, 14, 10, 4, 13, 21, 23, 18, 22, 20, 2, 3, 17, 16, 19, 1, 15, 12, 6, 5, 7, 9, 8]

    sample_mask = gu.remove_redundant_samples( sample_mask, samples_present )

    # splits = [660, 680, 765, 785, 890, 910, 1340, 1390, 1425, 1445, 1480, 1500, 1625, 1675, 1730, 1750, 2300, 2370, 2950, 2970, 3100, 3500]
    splits = [868, 885, 965, 980]

    feature = []

    range_mask = np.where( (data[0] >= splits[0]) & (data[0] < splits[1]) )

    for i in range( len( data[1] ) ):

        feature.append( gu.integral_2( data[0][range_mask], data[1][i][range_mask], 0, len( data[0][range_mask] ) - 1, data[1][i][range_mask].min() ) )

    features = np.array( feature )[:, np.newaxis]

    feature_names = ["FTIR_{}-{}".format( splits[0], splits[1] )]

    for s in range( 2, len( splits ) - 1, 2 ):

        feature = []

        range_mask = np.where( (data[0] >= splits[s]) & (data[0] < splits[s + 1]) )

        for i in range( len( data[1] ) ):

            feature.append( gu.integral_2( data[0][range_mask], data[1][i][range_mask], 0, len( data[0][range_mask] ) - 1, data[1][i][range_mask].min() ) )

        features = np.hstack( (features, np.array( feature )[:, np.newaxis]) )

        feature_names.append( "FTIR_{}-{}".format( splits[s], splits[s + 1] ) )

    feature_selection = [i for i in range( len( feature_names ) )]

    features = features[:, feature_selection]
    feature_names = list( np.array( feature_names )[feature_selection] )

    specimen_mask = gu.produce_mask( sample_array, sample_mask )

    features = features[specimen_mask, :]
    file_data_mask = np.array( file_data )[specimen_mask] # Orders file data according to specimen mask.
    file_data_mask = [[int( f[0] ), int( f[1] ), f[2], f[3]] for f in file_data_mask] # Converts file data so that f[0], etc. are integers and not np.str.
    sample_array = sample_array[specimen_mask]

    feature_weights = [1 for i in range( len( feature_selection ) )]

    gu.normalise_features( features, feature_weights )

    mean_features = gu.extract_mean_features( features, sample_array, sample_mask )
    mean_feature_names = feature_names.copy()

    std_of_features = gu.extract_std_of_features( features, sample_array, sample_mask )

    distance_matrix = gu.distance_matrix_from_features( features )

    mean_distance_matrix = gu.distance_matrix_from_features( mean_features )

    mean_features_plus_sample_mask = np.hstack( (np.array( sample_mask )[:, np.newaxis], mean_features) )

    df = gu.array_with_column_titles_to_df( mean_features_plus_sample_mask, ["sample"] + mean_feature_names )

    df.to_csv( output_directory + "FTIR/Integral_Analysis/Mean_Features.csv" )

    std_of_features_plus_sample_mask = np.hstack( (np.array( sample_mask )[:, np.newaxis], std_of_features) )

    df = gu.array_with_column_titles_to_df( std_of_features_plus_sample_mask, ["sample"] + mean_feature_names )

    df.to_csv( output_directory + "FTIR/Integral_Analysis/Std_of_Features.csv" )

    df = gu.array_with_column_titles_and_label_titles_to_df( mean_distance_matrix, sample_mask, sample_mask )

    df.to_csv( output_directory + "FTIR/Integral_Analysis/Distance_Matrix.csv" )

    if plot_specimen_distance_matrix:

        gu.plot_distance_matrix( output_directory, distance_matrix, [f[2] for f in file_data_mask], specimen = True, file_data = file_data_mask, sample_mask = sample_mask, subdirectory = "FTIR/Integral_Analysis/", title = "Specimen_Distance_Matrix.pdf" )

    if plot_mean_distance_matrix:

        gu.plot_distance_matrix( output_directory, mean_distance_matrix, [resin_data.loc[i]["Label"] for i in sample_mask], subdirectory = "FTIR/Integral_Analysis/", title = "Means_Distance_Matrix.pdf" )

    if plot_specimen_dendrogram:

        gu.plot_dendrogram( output_directory, distance_matrix, [f[2] for f in file_data_mask], specimen = True, subdirectory = "FTIR/Integral_Analysis/", title = "Specimen_Dendrogram.pdf" )

    if plot_mean_dendrogram:

        gu.plot_dendrogram( output_directory, mean_distance_matrix, [resin_data.loc[i]["Label"] for i in sample_mask], subdirectory = "FTIR/Integral_Analysis/", title = "Means_Dendrogram.pdf" )

def component_analysis( directory, output_directory, file_data, data ):

    plot_specimen_bars = False
    plot_mean_bars = False
    plot_specimen_features = True
    plot_mean_features = True

    resin_data = gu.get_list_of_resins_data( directory )

    sample, sample_array, samples_present, samples_present_array = gu.sample_data_from_file_data( file_data )

    sample_mask = [11, 14, 10, 4, 13, 21, 23, 18, 22, 20, 2, 3, 17, 16, 19, 1, 15, 12, 6, 5, 7, 9, 8]
    # sample_mask = [11, 14, 10, 4, 13, 21, 23, 18, 22, 20, 2, 3, 17, 16, 19, 1, 15, 12, 6, 5, 7, 9, 8, 401, 402, 403, 404, 405, 406, 407, 408, 409, 410, 411, 412, 413, 414, 415, 416]
    # sample_mask = [11, 14, 10, 4, 13, 21, 23, 18, 22, 20, 2, 3, 17, 16, 19, 1, 15, 12, 6, 5, 7, 9, 8, 500, 502, 504, 506, 508, 510, 512, 514, 516]
    # sample_mask = [500, 502, 504, 506, 508, 510, 512, 514, 516]

    sample_mask = gu.remove_redundant_samples( sample_mask, samples_present )

    pp_integrals = [835, 845, 993, 1005, 2950, 2970]

    features = []

    for s in range( 0, len( pp_integrals ) - 1, 2 ):

        feature = []

        range_mask = np.where( (data[0] >= pp_integrals[s]) & (data[0] < pp_integrals[s + 1]) )

        for i in range( len( data[1] ) ):

            integral = gu.integral_2( data[0][range_mask], data[1][i][range_mask], 0, len( data[0][range_mask] ) - 1, data[1][i][range_mask].min() )

            if integral < 0:

                integral = 0

            feature.append( integral )

        features.append( feature )

    features_array = np.array( features[0] )[:, np.newaxis]

    for i in range( 1, len( features ) ):

        features_array = np.hstack( (features_array, np.array( features[i] )[:, np.newaxis]) )

    gu.normalise_features( features_array )

    mean_integral = np.mean( features_array, 1 )[:, np.newaxis]
    feature_names = ["PP Content"]

    irgafos_integrals = [770, 785, 820, 830, 850, 860, 1075, 1090, 1485, 1495]

    features = []

    for s in range( 0, len( irgafos_integrals ) - 1, 2 ):

        feature = []

        range_mask = np.where( (data[0] >= irgafos_integrals[s]) & (data[0] < irgafos_integrals[s + 1]) )

        for i in range( len( data[1] ) ):

            integral = gu.integral_2( data[0][range_mask], data[1][i][range_mask], 0, len( data[0][range_mask] ) - 1, data[1][i][range_mask].min() )

            if integral < 0:

                integral = 0

            feature.append( integral )

        features.append( feature )

    features_array = np.array( features[0] )[:, np.newaxis]

    for i in range( 1, len( features ) ):

        features_array = np.hstack( (features_array, np.array( features[i] )[:, np.newaxis]) )

    gu.normalise_features( features_array )

    mean_integral = np.hstack( (mean_integral, np.mean( features_array, 1 )[:, np.newaxis]) )
    feature_names.append( "Irgafos Content" )

    CaCO3_integrals = [868, 885]

    features = []

    for s in range( 0, len( CaCO3_integrals ) - 1, 2 ):

        feature = []

        range_mask = np.where( (data[0] >= CaCO3_integrals[s]) & (data[0] < CaCO3_integrals[s + 1]) )

        for i in range( len( data[1] ) ):

            integral = gu.integral_2( data[0][range_mask], data[1][i][range_mask], 0, len( data[0][range_mask] ) - 1, data[1][i][range_mask].min() )

            if integral < 0:

                integral = 0

            feature.append( integral )

        features.append( feature )

    features_array = np.array( features[0] )[:, np.newaxis]

    for i in range( 1, len( features ) ):

        features_array = np.hstack( (features_array, np.array( features[i] )[:, np.newaxis]) )

    gu.normalise_features( features_array )

    mean_integral = np.hstack( (mean_integral, np.mean( features_array, 1 )[:, np.newaxis]) )
    feature_names.append( "CaCO3 Content" )

    vinyl_integrals = [906, 912]

    features = []

    for s in range( 0, len( vinyl_integrals ) - 1, 2 ):

        feature = []

        range_mask = np.where( (data[0] >= vinyl_integrals[s]) & (data[0] < vinyl_integrals[s + 1]) )

        for i in range( len( data[1] ) ):

            integral = gu.integral_2( data[0][range_mask], data[1][i][range_mask], 0, len( data[0][range_mask] ) - 1, data[1][i][range_mask].min() )

            if integral < 0:

                integral = 0

            feature.append( integral )

        features.append( feature )

    features_array = np.array( features[0] )[:, np.newaxis]

    for i in range( 1, len( features ) ):

        features_array = np.hstack( (features_array, np.array( features[i] )[:, np.newaxis]) )

    gu.normalise_features( features_array )

    mean_integral = np.hstack( (mean_integral, np.mean( features_array, 1 )[:, np.newaxis]) )
    feature_names.append( "Vinyl Content" )

    pet_integrals = [1015, 1025, 1095, 1105, 1110, 1125, 1405, 1415, 1715, 1730]

    features = []

    for s in range( 0, len( pet_integrals ) - 1, 2 ):

        feature = []

        range_mask = np.where( (data[0] >= pet_integrals[s]) & (data[0] < pet_integrals[s + 1]) )

        for i in range( len( data[1] ) ):

            integral = gu.integral_2( data[0][range_mask], data[1][i][range_mask], 0, len( data[0][range_mask] ) - 1, data[1][i][range_mask].min() )

            if integral < 0:

                integral = 0

            feature.append( integral )

        features.append( feature )

    features_array = np.array( features[0] )[:, np.newaxis]

    for i in range( 1, len( features ) ):

        features_array = np.hstack( (features_array, np.array( features[i] )[:, np.newaxis]) )

    gu.normalise_features( features_array )

    mean_integral = np.hstack( (mean_integral, np.mean( features_array, 1 )[:, np.newaxis]) )
    feature_names.append( "PET Content" )

    pa_integrals = [1635, 1650, 3280, 3320]

    features = []

    for s in range( 0, len( pa_integrals ) - 1, 2 ):

        feature = []

        range_mask = np.where( (data[0] >= pa_integrals[s]) & (data[0] < pa_integrals[s + 1]) )

        for i in range( len( data[1] ) ):

            integral = gu.integral_2( data[0][range_mask], data[1][i][range_mask], 0, len( data[0][range_mask] ) - 1, data[1][i][range_mask].min() )

            if integral < 0:

                integral = 0

            feature.append( integral )

        features.append( feature )

    features_array = np.array( features[0] )[:, np.newaxis]

    for i in range( 1, len( features ) ):

        features_array = np.hstack( (features_array, np.array( features[i] )[:, np.newaxis]) )

    gu.normalise_features( features_array )

    mean_integral = np.hstack( (mean_integral, np.mean( features_array, 1 )[:, np.newaxis]) )
    feature_names.append( "PA Content" )

    cast_integrals = [1570, 1585]

    features = []

    for s in range( 0, len( cast_integrals ) - 1, 2 ):

        feature = []

        range_mask = np.where( (data[0] >= cast_integrals[s]) & (data[0] < cast_integrals[s + 1]) )

        for i in range( len( data[1] ) ):

            integral = gu.integral_2( data[0][range_mask], data[1][i][range_mask], 0, len( data[0][range_mask] ) - 1, data[1][i][range_mask].min() )

            if integral < 0:

                integral = 0

            feature.append( integral )

        features.append( feature )

    features_array = np.array( features[0] )[:, np.newaxis]

    for i in range( 1, len( features ) ):

        features_array = np.hstack( (features_array, np.array( features[i] )[:, np.newaxis]) )

    gu.normalise_features( features_array )

    mean_integral = np.hstack( (mean_integral, np.mean( features_array, 1 )[:, np.newaxis]) )
    feature_names.append( "CaSt Content" )

    feature_selection = [i for i in range( len( feature_names ) )]

    mean_integral = mean_integral[:, feature_selection]
    feature_names = list( np.array( feature_names )[feature_selection] )

    mean_features_unnormalised = gu.extract_mean_features( mean_integral, sample_array, samples_present )
    mean_feature_names = feature_names.copy()

    mean_features_unnormalised_plus_sample_mask = np.hstack( (samples_present_array[:, np.newaxis], mean_features_unnormalised) )

    df = gu.array_with_column_titles_to_df( mean_features_unnormalised_plus_sample_mask, ["sample"] + mean_feature_names )

    df.to_csv( output_directory + "FTIR/Component_Analysis/Features/Mean_Features_Unnormalised.csv" )

    std_of_features = gu.extract_std_of_features( mean_integral, sample_array, samples_present )

    std_of_features_plus_sample_mask = np.hstack( (samples_present_array[:, np.newaxis], std_of_features) )

    df = gu.array_with_column_titles_to_df( std_of_features_plus_sample_mask, ["sample"] + mean_feature_names )

    df.to_csv( output_directory + "FTIR/Component_Analysis/Features/Std_of_Features_Unnormalised.csv" )

    specimen_mask = gu.produce_mask( sample_array, sample_mask )

    mean_integral = mean_integral[specimen_mask, :]
    file_data_mask = np.array( file_data )[specimen_mask] # Orders file data according to specimen mask.
    file_data_mask = [[int( f[0] ), int( f[1] ), f[2], f[3]] for f in file_data_mask] # Converts file data so that f[0], etc. are integers and not np.str.
    sample_array = sample_array[specimen_mask]

    mean_features = gu.extract_mean_features( mean_integral, sample_array, sample_mask )
    mean_feature_names = feature_names.copy()

    std_of_features = gu.extract_std_of_features( mean_integral, sample_array, sample_mask )

    mean_features_plus_sample_mask = np.hstack( (np.array( sample_mask )[:, np.newaxis], mean_features) )

    df = gu.array_with_column_titles_to_df( mean_features_plus_sample_mask, ["sample"] + mean_feature_names )

    df.to_csv( output_directory + "FTIR/Component_Analysis/Features/Mean_Features.csv" )

    std_of_features_plus_sample_mask = np.hstack( (np.array( sample_mask )[:, np.newaxis], std_of_features) )

    df = gu.array_with_column_titles_to_df( std_of_features_plus_sample_mask, ["sample"] + mean_feature_names )

    df.to_csv( output_directory + "FTIR/Component_Analysis/Features/Std_of_Features.csv" )

    if plot_specimen_bars:

        for i in range( len( mean_integral[0] ) ):

            gu.plot_barchart_of_feature( mean_integral[:, i], [f[2] for f in file_data_mask], colour = True, colour_mask = sample_array, filename = output_directory + "FTIR/Component_Analysis/Feature_Bars/Specimen/" + feature_names[i] + ".pdf", savefig = True )

    if plot_mean_bars:

        for i in range( len( mean_features[0] ) ):

            gu.plot_barchart_of_feature( mean_features[:, i], [resin_data.loc[i]["Label"] for i in sample_mask], errorbars = True, std = std_of_features[:, i], colour = True, colour_mask = sample_mask, filename = output_directory + "FTIR/Component_Analysis/Feature_Bars/Mean/" + mean_feature_names[i] + ".pdf", savefig = True )

    if plot_specimen_features:

        gu.plot_features( output_directory, mean_integral, feature_names, [f[2] for f in file_data_mask], specimen = True, subdirectory = "FTIR/Component_Analysis/Features/", title = "Specimen_Features.pdf" )

    if plot_mean_features:

        gu.plot_features( output_directory, mean_features, mean_feature_names, [resin_data.loc[i]["Label"] for i in sample_mask], subdirectory = "FTIR/Component_Analysis/Features/", title = "Means_Features.pdf" )

    import matplotlib.pyplot as plt

    # plt.plot( [0, 2, 4, 6, 8, 10, 12, 14, 16], mean_features, 'o' )

    # model = np.poly1d( np.polyfit( [0, 2, 4, 6, 8, 10, 12, 14, 16], mean_features[23:, 0], 2) )
    #
    # polyline = np.linspace( 0, 16, 100 )
    #
    # plt.scatter( [0, 2, 4, 6, 8, 10, 12, 14, 16], mean_features[23:, 0] )
    #
    # colours = gu.read_list_of_colours( directory )
    # scatter_colours = [colours[sample_mask[i]] for i in range( 23 )]
    #
    # plt.scatter( [(model - mean_features[i, 0]).roots[1] for i in range( 23 )], mean_features[:23, 0], color = scatter_colours, s = 100 )
    #
    # plt.plot(polyline, model( polyline ) )
    #
    # plt.xlabel( "Known concentration of PP (%)", fontsize = 13 )
    # plt.ylabel( "Model value", fontsize = 13 )
    #
    # plt.show()
    #
    # plt.close()

def specimen_feature_correlation( directory, output_directory, file_data ):

    df = pd.read_csv( output_directory + "FTIR/Features/Specimen_Features.csv" )

    df.drop( columns = df.columns[0], inplace = True )

    # feature_1 = "FTIR_667"
    # feature_2 = "FTIR_2349"
    #
    # pearson, _ = pearsonr( df[feature_1].to_list(), df[feature_2].to_list() )
    #
    # print( pearson )
    #
    # gu.plot_scatterplot_of_two_specimen_features_with_hover_annotation( df[feature_1].to_list(), df[feature_2].to_list(), file_data )

    # for i in range( len( df.columns ) ):
    #
    #     feature_1 = df.columns[i]
    #
    #     for j in range( i + 1, len( df.columns ) ):
    #
    #         feature_2 = df.columns[j]
    #
    #         pearson, _ = pearsonr( df[feature_1].to_list(), df[feature_2].to_list() )
    #
    #         if abs( pearson ) > 0.75:
    #
    #             print( pearson, feature_1, feature_2 )
    #
    #             gu.plot_scatterplot_of_two_specimen_features_with_hover_annotation( df[feature_1].to_list(), df[feature_2].to_list(), file_data )

    feature_1 = "FTIR_663"

    for j in range( len( df.columns ) ):

        feature_2 = df.columns[j]

        pearson, _ = pearsonr( df[feature_1].to_list(), df[feature_2].to_list() )

        if abs( pearson ) > 0.75:

            print( pearson, feature_1, feature_2 )

            gu.plot_scatterplot_of_two_specimen_features_with_hover_annotation( df[feature_1].to_list(), df[feature_2].to_list(), file_data )

def correlation_with_external_data( directory, output_directory, file_data, data ):
    '''Find correlations with external data.'''

    resin_data = gu.get_list_of_resins_data( directory )

    external_data = pd.read_csv( output_directory + "DSC/Output/Features/Mean_Features.csv" )

    external_data.drop( columns = [external_data.columns[0]], inplace = True )

    samples_present_external_data = external_data.iloc[:, 0].tolist()

    sample, sample_array, samples_present, samples_present_array = gu.sample_data_from_file_data( file_data )

    sample_mask = samples_present

    sample_mask = gu.remove_redundant_samples( sample_mask, samples_present )

    sample_mask = gu.remove_redundant_samples( sample_mask, samples_present_external_data )

    feature_list = []

    for s in sample_mask:

        feature_list.append( external_data.iloc[samples_present_external_data.index( s ), 1] )

    feature = np.array( feature_list )

    max_covalence = 0
    max_pearson = 0
    index_covalence = -1
    index_pearson = -1

    for i in range( len( data[0] ) ): # Loop over wavenumbers.

        wavenumber_values = []

        for j in range( len( data[2] ) ):

            wavenumber_values.append( data[2][j][i] )

        wavenumber_values_array = np.array( wavenumber_values )

        covalence = np.cov( feature, wavenumber_values_array )[0][1]
        pearson, _ = pearsonr( feature, wavenumber_values_array )

        if max_pearson < pearson:

            max_pearson = pearson
            index_pearson = i

        if max_covalence < covalence:

            max_covalence = covalence
            index_covalence = i

    print( max_pearson, index_pearson )
    print( max_covalence, index_covalence )

    wavenumber_values = []

    for j in range( len( data[2] ) ):

        wavenumber_values.append( data[2][j][index_pearson] )

    wavenumber_values_array = np.array( wavenumber_values )

    gu.plot_scatterplot_of_two_features( feature, wavenumber_values_array, sample_mask, [resin_data.loc[i]["Label"] for i in sample_mask] )

def normalisation_experimentation( directory, output_directory, file_data, data ):
    '''Experimenting to find best preprocessing for PCA.'''

    for i in range( len( data[1] ) ):

        # data[1][i] = (data[1][i] / data[1][i].max())
        data[1][i] = ((data[1][i] - data[1][i].min()) / (data[1][i].max() - data[1][i].min()))

    Preprocessing.compute_mean( output_directory, file_data, data )

    Preprocessing.read_mean( output_directory, data )

    first_derivative_data = compute_derivatives( data )
    second_derivative_data = compute_derivatives( first_derivative_data )
    third_derivative_data = compute_derivatives( second_derivative_data )

    m_peaks, m_peaks_array, y_peaks, y_peaks_array = extract_FTIR_features( output_directory, file_data, data, first_derivative_data, second_derivative_data, third_derivative_data, -0.00008, 0.0002, 6, 3400 )

    resin_data = gu.get_list_of_resins_data( directory )

    feature_names, features = gu.csv_to_df_to_array_and_column_titles( output_directory + "FTIR/Features/Specimen_Features.csv" )

    ###

    features = first_derivative_data[1]
    feature_names = [str( i ) for i in first_derivative_data[0]]

    ###

    sample, sample_array, samples_present, samples_present_array = gu.sample_data_from_file_data( file_data )

    sample_mask = [11, 14, 10, 4, 13, 21, 23, 18, 22, 20, 2, 3, 17, 16, 19, 1, 15, 12, 6, 5, 7, 9, 8, 24]

    sample_mask = gu.remove_redundant_samples( sample_mask, samples_present )

    feature_selection = [i for i in range( len( feature_names ) )]

    features = features[:, feature_selection]
    feature_names = list( np.array( feature_names )[feature_selection] )

    specimen_mask = gu.produce_mask( sample_array, sample_mask )

    features = features[specimen_mask, :]
    file_data_mask = np.array( file_data )[specimen_mask] # Orders file data according to specimen mask.
    file_data_mask = [[int( f[0] ), int( f[1] ), f[2], f[3]] for f in file_data_mask] # Converts file data so that f[0], etc. are integers and not np.str.
    sample_array = sample_array[specimen_mask]

    feature_weights = [1 for i in range( len( feature_selection ) )]

    # gu.normalise_features( features, feature_weights )

    mean_features = gu.extract_mean_features( features, sample_array, sample_mask )
    mean_feature_names = feature_names.copy()

    std_of_features = gu.extract_std_of_features( features, sample_array, sample_mask )

    distance_matrix = gu.distance_matrix_from_features( features )

    mean_distance_matrix = gu.distance_matrix_from_features( mean_features )

    mean_features_plus_sample_mask = np.hstack( (np.array( sample_mask )[:, np.newaxis], mean_features) )

    df = gu.array_with_column_titles_to_df( mean_features_plus_sample_mask, ["sample"] + mean_feature_names )

    df.to_csv( output_directory + "FTIR/Features/Mean_Features.csv" )

    std_of_features_plus_sample_mask = np.hstack( (np.array( sample_mask )[:, np.newaxis], std_of_features) )

    df = gu.array_with_column_titles_to_df( std_of_features_plus_sample_mask, ["sample"] + mean_feature_names )

    df.to_csv( output_directory + "FTIR/Features/Std_of_Features.csv" )

    df = gu.array_with_column_titles_and_label_titles_to_df( mean_distance_matrix, sample_mask, sample_mask )

    df.to_csv( output_directory + "FTIR/Features/Distance_Matrix.csv" )

    # gu.plot_features( output_directory, mean_features, [f[5:] for f in mean_feature_names], [resin_data.loc[i]["Label"] for i in sample_mask], subdirectory = "FTIR/Features/", title = "Means_Features.pdf" )

    features_df = gu.array_with_column_titles_and_label_titles_to_df( mean_features, [str( i ) for i in mean_feature_names], sample_mask )

    # features_df = StandardScaler().fit_transform( features_df )

    features_df = gu.array_with_column_titles_and_label_titles_to_df( features_df, [str( i ) for i in mean_feature_names], sample_mask )

    std_of_features_df = gu.array_with_column_titles_and_label_titles_to_df( std_of_features, [str( i ) for i in mean_feature_names], sample_mask )

    util2.pca( directory, features_df, std_of_features_df )

    # Default:
    # [0.0, 0.35883070579474546, 0.58859988292701, 0.7390626819327278, 0.8210366863949823, 0.8693373369051673, 0.904292376728703, 0.9304661729563468, 0.9513682016298718, 0.9652131579541247, 0.975691795171862]

    # No normalisation:
    # [0.0, 0.3524054219641353, 0.604358407390208, 0.7488330420945205, 0.8358651978695985, 0.8823789020545217, 0.9102495323025178, 0.9342043003316911, 0.954834274196833, 0.9690753972011436, 0.9790872682977163]

    # Max-min normalisation:
    # [0.0, 0.35975369836811844, 0.5878277260239937, 0.738444752490623, 0.8204529185163695, 0.8686913461725618, 0.9041059160564926, 0.9303498261009217, 0.9514856417411043, 0.9652947741566771, 0.9756082339872473]

    # Max-min normalisation, all in:
    # [0.0, 0.519424198453715, 0.7109133621521699, 0.8403923836430574, 0.9029294068088906, 0.9354452996364587, 0.9563529416662541, 0.9693791609314283, 0.9812628941776536, 0.987887135278509, 0.9918010241419203]

    # Max-min normalisation, all in, first derivative:
    # [0.0, 0.43051277477178207, 0.8298189845995849, 0.9090705223769208, 0.9442834667082102, 0.959899984490423, 0.9692436649075102, 0.9761112933461358, 0.9817540207900024, 0.9857847801988655, 0.9894756652385638]

def pp_percentage_model( directory, output_directory, file_data, data, name_appendage = "" ):

    blend_samples = [500, 502, 504, 506, 508, 510, 512, 514, 516] # Samples with known ratios of HDPE/PP.

    resin_data = gu.get_list_of_resins_data( directory, name_appendage )

    sample, sample_array, samples_present, samples_present_array = gu.sample_data_from_file_data( file_data )

    sample_mask = [11, 14, 10, 4, 13, 21, 23, 18, 22, 20, 2, 3, 17, 16, 19, 1, 15, 12, 6, 5, 7, 9, 8, 401, 402, 403, 404, 405, 406, 407, 408, 409, 410, 411, 412, 413, 414, 415, 416, 601, 602, 603, 604, 605, 701, 702, 703, 704, 705, 706, 707, 708, 709, 710, 711, 712, 500, 502, 504, 506, 508, 510, 512, 514, 516, 518]
    sample_mask = samples_present

    sample_mask = gu.remove_redundant_samples( sample_mask, samples_present )

    # Applying minmax normalisation.

    range_mask = np.where( (data[0] >= 3050) & (data[0] < 3100) )

    for i in range( len( data[1] ) ):

        mini = data[1][i][range_mask].min()

        data[1][i] = ((data[1][i] - mini) / (data[1][i].max() - mini))

    ######
    # Scaling data so that all spectra have a peak magnitude of 0.1 or less at 720.

    range_mask = np.where( (data[0] >= 712) & (data[0] < 726) )

    for i in range( len( data[1] ) ):

        value = data[1][i][range_mask].max()

        if value > 0.1:

            data[1][i] = data[1][i] * 0.1 / value

    # Integrating PP peaks.

    pp_integrals = [801, 815, 834, 848, 966, 980, 990, 1004, 1160, 1174, 1365, 1385] # [803, 813, 836, 846, 968, 978, 992, 1002, 1162, 1172, 1375, 1382] # Previous intervals used, now use slightly broader ones.
    feature_names = ["808", "841", "973", "997", "1167", "1377"]

    features = []

    for s in range( 0, len( pp_integrals ) - 1, 2 ):

        feature = []

        range_mask = np.where( (data[0] >= pp_integrals[s]) & (data[0] < pp_integrals[s + 1]) )

        for i in range( len( data[1] ) ):

            integral = gu.integral_2( data[0][range_mask], data[1][i][range_mask], 0, len( data[0][range_mask] ) - 1, data[1][i][range_mask].min() )

            if integral < 0:

                integral = 0

            feature.append( integral )

        features.append( feature )

    features_array = np.array( features[0] )[:, np.newaxis]

    for i in range( 1, len( features ) ):

        features_array = np.hstack( (features_array, np.array( features[i] )[:, np.newaxis]) )

    specimen_mask = gu.produce_mask( sample_array, sample_mask )

    features_array = features_array[specimen_mask, :]
    file_data_mask = np.array( file_data )[specimen_mask] # Orders file data according to specimen mask.
    file_data_mask = [[int( f[0] ), int( f[1] ), f[2], f[3]] for f in file_data_mask] # Converts file data so that f[0], etc. are integers and not np.str.
    sample_array = sample_array[specimen_mask]

    mean_features_without_extremities_pre_adjust = gu.extract_mean_features_2( features_array, sample_array, sample_mask )
    mean_features_without_extremities = mean_features_without_extremities_pre_adjust - mean_features_without_extremities_pre_adjust[sample_mask.index( 500 ), :]

    # mean_features = gu.extract_mean_features( features_array, sample_array, sample_mask )
    # mean_feature_names = feature_names.copy()
    #
    # std_of_features = gu.extract_std_of_features( features_array, sample_array, sample_mask )
    #
    # mean_features_plus_sample_mask = np.hstack( (np.array( sample_mask )[:, np.newaxis], mean_features) )
    #
    # df = gu.array_with_column_titles_to_df( mean_features_plus_sample_mask, ["sample"] + mean_feature_names )
    #
    # df.to_csv( output_directory + "FTIR/Sandbox/PP_Percentage_Analysis/Features/Mean_Features.csv" )
    #
    # std_of_features_plus_sample_mask = np.hstack( (np.array( sample_mask )[:, np.newaxis], std_of_features) )
    #
    # df = gu.array_with_column_titles_to_df( std_of_features_plus_sample_mask, ["sample"] + mean_feature_names )
    #
    # df.to_csv( output_directory + "FTIR/Sandbox/PP_Percentage_Analysis/Features/Std_of_Features.csv" )
    #
    # for i in range( len( features_array[0] ) ):
    #
    #     gu.plot_barchart_of_feature( features_array[:, i], [f[2] for f in file_data_mask], colour = True, colour_mask = sample_array, filename = output_directory + "FTIR/Sandbox/PP_Percentage_Analysis/Feature_Bars/Specimen/" + feature_names[i] + ".pdf", savefig = True )
    #
    # for i in range( len( mean_features[0] ) ):
    #
    #     gu.plot_barchart_of_feature( mean_features[:, i], [resin_data.loc[i]["Label"] for i in sample_mask], errorbars = True, std = std_of_features[:, i], colour = True, colour_mask = sample_mask, filename = output_directory + "FTIR/Sandbox/PP_Percentage_Analysis/Feature_Bars/Mean/" + mean_feature_names[i] + ".pdf", savefig = True )
    #
    # gu.plot_features( output_directory, features_array, feature_names, [f[2] for f in file_data_mask], specimen = True, subdirectory = "FTIR/Sandbox/PP_Percentage_Analysis/Features/", title = "Specimen_Features.pdf" )
    #
    # gu.plot_features( output_directory, mean_features, mean_feature_names, [resin_data.loc[i]["Label"] for i in sample_mask], subdirectory = "FTIR/Sandbox/PP_Percentage_Analysis/Features/", title = "Means_Features.pdf" )

    def func( x, a ):

        return a * x

    def positive_definite_and_ton_limit( x ):

        if x < 0:

            return 0.0

        elif x > 100:

            return 100.0

        else:

            return x

    vfunc = np.vectorize( positive_definite_and_ton_limit )

    popts = []

    for i in range( len( features_array[0] ) ):

        popt, _ = curve_fit( func, [0, 2, 4, 6, 8, 10, 12, 14, 16], mean_features_without_extremities[[sample_mask.index( j ) for j in blend_samples], i] )

        popts.append( popt )

        mean_features_without_extremities[:, i] = mean_features_without_extremities[:, i] / popt[0]

        features_array[:, i] = vfunc( (features_array[:, i] - mean_features_without_extremities_pre_adjust[sample_mask.index( 500 ), i]) / popt[0] )

    df = gu.array_with_column_titles_and_label_titles_to_df( features_array, feature_names, [file_data_mask[i][0] for i in range( len( file_data_mask ) )] )

    df.to_csv( output_directory + "FTIR/Sandbox/PP_Percentage_Analysis/Features/PP_Predictions" + name_appendage + ".csv", float_format = "%.3f" )

    final_pp_prediction = []

    for i in sample_mask:

        specimen_mask = gu.produce_mask( sample_array, [i] )

        features_array_copy = features_array[specimen_mask, :]

        specimen_predictions = []

        for j in range( len( features_array_copy ) ):

            specimen_predictions.append( (features_array_copy[j].sum() - features_array_copy[j].max() - features_array_copy[j].min()) / (len( features_array_copy[j] ) - 2) )

        if len( specimen_predictions ) >= 5:

            array = np.array( specimen_predictions )

            final_pp_prediction.append( (array.sum() - array.max() - array.min()) / (len( array ) - 2) )

        else:

            array = np.array( specimen_predictions )

            final_pp_prediction.append( array.mean() )

        # print( i, final_pp_prediction[-1] )

    pearson, _ = pearsonr( [0, 2, 4, 6, 8, 10, 12, 14, 16], np.array( final_pp_prediction )[[sample_mask.index( j ) for j in blend_samples]] )

    # print( "R squared value of PP prediction is ", pearson * pearson )

    # y_pos = np.arange( 6 ) * 9
    #
    # for i in range( 9 ):
    #
    #     plt.bar( y_pos + i, mean_features_without_extremities[sample_mask.index( blend_samples[i] ), :], align = 'center', alpha = 0.5 )
    #
    # plt.tight_layout()
    #
    # plt.show()
    #
    # plt.close()

    ###### Miles's approach

    feature = []

    range_mask_1 = np.where( (data[0] >= 3090) & (data[0] < 3110) )
    range_mask_2 = np.where( (data[0] >= 1374) & (data[0] < 1379) )
    range_mask_3 = np.where( (data[0] >= 717) & (data[0] < 722) )

    for i in range( len( data[1] ) ):

        value_1 = data[1][i][range_mask_1].min()
        value_2 = data[1][i][range_mask_2].max()
        value_3 = data[1][i][range_mask_3].max()

        feature.append( (value_2 - value_1) / (value_3 + value_2 - value_1 - value_1) )

    features_array = np.array( feature )[:, np.newaxis]

    sample, sample_array, samples_present, samples_present_array = gu.sample_data_from_file_data( file_data )

    specimen_mask = gu.produce_mask( sample_array, sample_mask )

    features_array = features_array[specimen_mask, :]
    sample_array = sample_array[specimen_mask]

    mean_features_without_extremities_pre_adjust = gu.extract_mean_features_2( features_array, sample_array, sample_mask )
    mean_features_without_extremities = mean_features_without_extremities_pre_adjust - mean_features_without_extremities_pre_adjust[sample_mask.index( 500 ), :]

    popt, _ = curve_fit( func, [0, 2, 4, 6, 8, 10, 12, 14, 16], mean_features_without_extremities[[sample_mask.index( j ) for j in blend_samples], 0] )

    mean_features_without_extremities[:, 0] = mean_features_without_extremities[:, 0] / popt[0]

    features_array[:, 0] = vfunc( (features_array[:, 0] - mean_features_without_extremities_pre_adjust[sample_mask.index( 500 ), 0]) / popt[0] )

    final_pp_prediction_2 = []

    for i in sample_mask:

        specimen_mask = gu.produce_mask( sample_array, [i] )

        features_array_copy = features_array[specimen_mask, :]

        if len( features_array_copy[:, 0] ) >= 5:

            array = features_array_copy[:, 0]

            final_pp_prediction_2.append( (array.sum() - array.max() - array.min()) / (len( array ) - 2) )

        else:

            array = features_array_copy[:, 0]

            final_pp_prediction_2.append( array.mean() )

        # print( i, final_pp_prediction_2[-1] )

    pearson, _ = pearsonr( [0, 2, 4, 6, 8, 10, 12, 14, 16], np.array( final_pp_prediction_2 )[[sample_mask.index( j ) for j in blend_samples]] )

    # print( "R squared value of PP prediction is ", pearson * pearson )

def crystallinity_determination( directory, output_directory, file_data, data ):

    resin_data = gu.get_list_of_resins_data( directory )

    sample, sample_array, samples_present, samples_present_array = gu.sample_data_from_file_data( file_data )

    sample_mask = [11, 14, 10, 4, 13, 21, 18, 22, 20, 2, 3, 17, 16, 19, 1, 15, 12, 6, 5, 7, 9, 8]

    sample_mask = gu.remove_redundant_samples( sample_mask, samples_present )

    features = []

    feature = []

    range_mask_1 = np.where( (data[0] >= 755) & (data[0] < 765) )
    range_mask_2 = np.where( (data[0] >= 715) & (data[0] < 725) )
    range_mask_3 = np.where( (data[0] >= 725) & (data[0] < 735) )

    for i in range( len( data[1] ) ):

        i_a = data[1][i][range_mask_3].max() - data[1][i][range_mask_1].min()
        i_b = data[1][i][range_mask_2].max() - data[1][i][range_mask_1].min()

        cryst = (1 - (i_a / i_b) / 1.233) / (1 + i_a / i_b) * 100

        feature.append( cryst )

    features.append( feature )

    feature = []

    for i in range( len( data[1] ) ):

        i_a = data[1][i][range_mask_2].max() - data[1][i][range_mask_3].max()

        feature.append( i_a )

    features.append( feature )

    feature = []

    range_mask_1 = np.where( (data[0] >= 1500) & (data[0] < 1525) )
    range_mask_2 = np.where( (data[0] >= 1460) & (data[0] < 1468) )
    range_mask_3 = np.where( (data[0] >= 1470) & (data[0] < 1480) )

    for i in range( len( data[1] ) ):

        i_a = data[1][i][range_mask_3].max() - data[1][i][range_mask_1].min()
        i_b = data[1][i][range_mask_2].max() - data[1][i][range_mask_1].min()

        cryst = (1 - (i_a / i_b) / 1.233) / (1 + i_a / i_b) * 100

        feature.append( cryst )

    features.append( feature )

    feature_names = ["FTIR_Cryst1", "FTIR_Cryst2", "FTIR_Cryst3"]

    features_array = np.array( features[0] )[:, np.newaxis]

    features_array = np.hstack( (features_array, np.array( features[1] )[:, np.newaxis]) )
    features_array = np.hstack( (features_array, np.array( features[2] )[:, np.newaxis]) )

    specimen_mask = gu.produce_mask( sample_array, sample_mask )

    features_array = features_array[specimen_mask, :]
    file_data_mask = np.array( file_data )[specimen_mask] # Orders file data according to specimen mask.
    file_data_mask = [[int( f[0] ), int( f[1] ), f[2], f[3]] for f in file_data_mask] # Converts file data so that f[0], etc. are integers and not np.str.
    sample_array = sample_array[specimen_mask]

    gu.normalise_features( features_array )

    features = features_array

    mean_features = gu.extract_mean_features( features, sample_array, sample_mask )
    mean_feature_names = feature_names.copy()

    std_of_features = gu.extract_std_of_features( features, sample_array, sample_mask )

    mean_distance_matrix = gu.distance_matrix_from_features( mean_features )

    mean_features_plus_sample_mask = np.hstack( (np.array( sample_mask )[:, np.newaxis], mean_features) )

    df = gu.array_with_column_titles_to_df( mean_features_plus_sample_mask, ["sample"] + mean_feature_names )

    df.to_csv( output_directory + "FTIR/Sandbox/Crystallinity/Mean_Features.csv" )

    std_of_features_plus_sample_mask = np.hstack( (np.array( sample_mask )[:, np.newaxis], std_of_features) )

    df = gu.array_with_column_titles_to_df( std_of_features_plus_sample_mask, ["sample"] + mean_feature_names )

    df.to_csv( output_directory + "FTIR/Sandbox/Crystallinity/Std_of_Features.csv" )

    df = gu.array_with_column_titles_and_label_titles_to_df( mean_distance_matrix, sample_mask, sample_mask )

    df.to_csv( output_directory + "FTIR/Sandbox/Crystallinity/Distance_Matrix.csv" )

    gu.plot_features( output_directory, features, feature_names, [f[2] for f in file_data_mask], specimen = True, subdirectory = "FTIR/Sandbox/Crystallinity/", title = "Specimen_Features.pdf" )

    gu.plot_features( output_directory, mean_features, mean_feature_names, [resin_data.loc[i]["Label"] for i in sample_mask], subdirectory = "FTIR/Sandbox/Crystallinity/", title = "Means_Features.pdf" )

def sandbox( directory, output_directory, file_data, data, first_derivative_data, second_derivative_data, third_derivative_data ):

    perform_integral_analysis = False # Perform a second kind of analysis on the spectrums focusing on integrals.

    perform_component_analysis = True

    perform_specimen_feature_correlation = False

    perform_correlation_with_external_data = False # Try to find correlations between external data and FTIR data.

    perform_normalisation_experimentation = False

    perform_pp_percentage_model = False

    perform_crystallinity_determination = False

    if perform_integral_analysis:

        integral_analysis( directory, output_directory, file_data, data )

    if perform_component_analysis:

        component_analysis( directory, output_directory, file_data, data )

    if perform_specimen_feature_correlation:

        specimen_feature_correlation( directory, output_directory, file_data )

    if perform_correlation_with_external_data:

        correlation_with_external_data( directory, output_directory, file_data, data )

    if perform_normalisation_experimentation:

        normalisation_experimentation( directory, output_directory, file_data, data )

    if perform_pp_percentage_model:

        pp_percentage_model( directory, output_directory, file_data, data )

    if perform_crystallinity_determination:

        crystallinity_determination( directory, output_directory, file_data, data )



# Imports.

from . import Preprocessing
from . import Utilities as util
from . import TGA_plotting
from .. import Global_Utilities as gu

# Main function definition.

def TGA_Analysis_Main( ip ):

    file_data, data = [], []

    if ip.read_files:

        file_data, data = Preprocessing.read_files_and_preprocess( ip.directory, ip.data_directory, ip.merge_groups )

    if ip.write_csv:

        Preprocessing.write_csv( ip.output_directory, file_data, data )

    if ip.read_csv:

        file_data, data = Preprocessing.read_csv( ip.directory, ip.output_directory, ip.merge_groups )

    # Data of form [[Time], Temp, [Weight (mg)], [Weight %], [Derivative Weight %/Temp]]. Data of type lists of numpys.

    gu.print_files_read( len( data[3] ), 53 )

    if ip.remove_files:

        Preprocessing.remove_files( file_data, data, "fs" )

    if ip.compute_mean:

        Preprocessing.compute_mean( ip.output_directory, file_data, data )

    if ip.read_mean:

        Preprocessing.read_mean( ip.output_directory, data )

    # Data now of form [..., [m]]. Data of type lists of numpys.

    if ip.derivative:

        first_derivative_data = util.compute_derivatives( data, width = 50 )
        second_derivative_data = util.compute_derivatives( first_derivative_data )

    if ip.extract_features:

        util.extract_TGA_features( ip.output_directory, file_data, data, first_derivative_data, second_derivative_data )

    if ip.read_and_analyse_features:

        util.read_and_analyse_TGA_features( ip, file_data )

    if ip.plot_data:

        TGA_plotting.plot_data( ip, file_data, data, first_derivative_data, second_derivative_data )

    if ip.sandbox:

        util.sandbox( ip.directory, ip.output_directory, file_data, data, first_derivative_data, second_derivative_data )



# Imports.

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib as mpl
import time

from .. import Global_Utilities as gu

# Function definitions.

def plot_data( ip, file_data, data, first_derivative_data, second_derivative_data, third_derivative_data, m_peaks, m_peaks_array, y_peaks, y_peaks_array, savefig = False ):

    # for i in range( len( data[1] ) ):
    #
    #     # data[1][i] = (data[1][i] / data[1][i].max())
    #     data[1][i] = ((data[1][i] - data[1][i].min()) / (data[1][i].max() - data[1][i].min()))
    #
    # for i in range( len( data[2] ) ):
    #
    #     # data[1][i] = (data[1][i] / data[1][i].max())
    #     data[2][i] = ((data[2][i] - data[2][i].min()) / (data[2][i].max() - data[2][i].min()))

    resin_data = gu.get_list_of_resins_data( ip.directory, ip.name_appendage )

    # For overall pipeline figure.

    # mpl.rcParams['lines.linewidth'] = 3 #4

    sample, sample_array, samples_present, samples_present_array = gu.sample_data_from_file_data( file_data )

    # samples_to_plot = samples_present
    samples_to_plot = [1, 2, 3, 4, 701]

    specimens = False
    all_specimens = True
    specimen_mask_by_index = [0, 1, 3]
    specimen_mask = []

    mean = True

    plot_mean_peaks = False
    plot_specimen_peaks = False

    deriv0 = True
    deriv1 = False
    deriv2 = False
    deriv3 = False

    split = True
    # splits = [int( data[0][len( data[0] ) - 1] ), 800, 1400, 1500, 2000, 2700, 3000, int( data[0][0] )]
    # splits = [600, 650, 700, 750, 800, 850, 900, 950, 1000]
    # splits = [1000, 1050, 1100, 1150, 1200, 1250, 1300, 1350, 1400, 1450, 1500, 1550, 1600, 1650, 1700, 1750, 1800, 1850, 1900, 1950, 2000]
    # splits = [2000, 2050, 2100, 2150, 2200, 2250, 2300, 2350, 2400, 2450, 2500, 2550, 2600, 2650, 2700, 2750, 2800, 2850, 2900, 2950, 3000]
    # splits = [3000, 3050, 3100, 3150, 3200, 3250, 3300, 3350, 3400, 3450, 3500, 3550, 3600, 3650, 3700, 3750, 3800, 3850, 3900, 3950, 4000]
    # splits = [3000, 4000]
    # splits = [600, 1000, 1400, 1800, 2200, 2600, 3000]
    splits = [800, 1000]

    if not split:

        splits = [int( data[0][len( data[0] ) - 1] ), int( data[0][0] )]

    if ip.shiny:

        samples_to_plot = ip.shiny_samples_to_plot

        if type( samples_to_plot ) == int:

            samples_to_plot = [samples_to_plot]

        specimen_mask = ip.shiny_specimens_to_plot

        if type( specimen_mask ) == int:

            specimen_mask = [specimen_mask]

        splits = ip.shiny_split
        mean = ip.shiny_mean
        specimens = ip.shiny_specimen
        all_specimens = False

    colours = gu.read_list_of_colours( ip.directory )

    shiny_de = []

    data_extraction_bool = False

    for s in range( len( splits ) - 1 ):

        data_extraction = []

        lower_bound, upper_bound = splits[s], splits[s + 1]

        for i in samples_to_plot:

            if plot_mean_peaks:

                index = np.where( samples_present_array == i )[0][0]

                for p in range( len( m_peaks ) ):

                    if m_peaks[p] > lower_bound and m_peaks[p] < upper_bound:

                        if m_peaks_array[index][p] > 0.0001:

                            plt.axvline( x = m_peaks[p], color = colours[samples_present[index]] )

            if specimens:

                mask = np.where( sample_array == i )[0]

                if not ip.shiny:

                    if all_specimens:

                        specimen_mask = [file_data[mask[j]][2] for j in range( len( mask ) )]

                    else:

                        specimen_mask = [file_data[mask[j]][2] for j in specimen_mask_by_index]

                for j in mask:

                    if file_data[j][2] in specimen_mask:

                        if plot_specimen_peaks:

                            for p in range( len( y_peaks ) ):

                                if y_peaks[p] > lower_bound and y_peaks[p] < upper_bound:

                                    if y_peaks_array[j][p] > 0.0001:

                                        plt.axvline( x = y_peaks[p] )

                        if deriv0:

                            wn_mask = np.where( (data[0] <= upper_bound) & (data[0] >= lower_bound) )[0]

                            plt.plot( data[0][wn_mask], data[1][j][wn_mask], label = file_data[j][2] )

                            shiny_de.append( data[0][wn_mask].tolist() )
                            shiny_de.append( data[1][j][wn_mask].tolist() )

                        if deriv1:

                            wn_mask = np.where( (first_derivative_data[0] <= upper_bound) & (first_derivative_data[0] >= lower_bound) )[0]

                            plt.plot( first_derivative_data[0][wn_mask], first_derivative_data[1][j][wn_mask], label = file_data[j][2] )

                        if deriv2:

                            wn_mask = np.where( (second_derivative_data[0] <= upper_bound) & (second_derivative_data[0] >= lower_bound) )[0]

                            plt.plot( second_derivative_data[0][wn_mask], second_derivative_data[1][j][wn_mask], label = file_data[j][2] )

                        if deriv3:

                            wn_mask = np.where( (third_derivative_data[0] <= upper_bound) & (third_derivative_data[0] >= lower_bound) )[0]

                            plt.plot( third_derivative_data[0][wn_mask], third_derivative_data[1][j][wn_mask], label = file_data[j][2] )

            if mean:

                index = np.where( samples_present_array == i )[0][0]

                if deriv0:

                    wn_mask = np.where( (data[0] <= upper_bound) & (data[0] >= lower_bound) )[0]

                    plt.plot( data[0][wn_mask], data[2][index][wn_mask], label = resin_data.loc[i]["Label"], color = colours[i] )

                    shiny_de.append( data[0][wn_mask].tolist() )
                    shiny_de.append( data[2][index][wn_mask].tolist() )

                if deriv1:

                    wn_mask = np.where( (first_derivative_data[0] <= upper_bound) & (first_derivative_data[0] >= lower_bound) )[0]

                    plt.plot( first_derivative_data[0][wn_mask], first_derivative_data[2][index][wn_mask], label = resin_data.loc[i]["Label"], color = colours[i] )

                if deriv2:

                    wn_mask = np.where( (second_derivative_data[0] <= upper_bound) & (second_derivative_data[0] >= lower_bound) )[0]

                    plt.plot( second_derivative_data[0][wn_mask], second_derivative_data[2][index][wn_mask], label = resin_data.loc[i]["Label"], color = colours[i] )

                if deriv3:

                    wn_mask = np.where( (third_derivative_data[0] <= upper_bound) & (third_derivative_data[0] >= lower_bound) )[0]

                    plt.plot( third_derivative_data[0][wn_mask], third_derivative_data[2][index][wn_mask], label = resin_data.loc[i]["Label"], color = colours[i] )

                    # data_extraction.append( third_derivative_data[0][wn_mask] )
                    # data_extraction.append( third_derivative_data[2][index][wn_mask] )

        if ip.shiny:

            return shiny_de

        plt.legend( ncol = 2, bbox_to_anchor = ( 1.05, 1 ), loc = 'upper left', borderaxespad = 0 )
        # plt.legend( ncol = 2, loc = 'upper right', borderaxespad = 0, fontsize = 20 )
        # plt.legend()

        plt.xlabel( "Wavenumber [cm\u207b\u00b9]", fontsize = 20 )
        plt.ylabel( "Absorbance", fontsize = 20 )
        plt.xticks( fontsize = 20 )
        plt.yticks( fontsize = 20 )

        ax = plt.gca()
        ax.invert_xaxis()

        plt.tight_layout()

        # For overall pipeline figure.

        # ax.get_legend().remove()
        # plt.xlabel( "" )
        # plt.ylabel( "" )
        # plt.tick_params( axis = 'x', which = 'both', bottom = False, top = False, labelbottom = False )
        # plt.tick_params( axis = 'y', which = 'both', left = False, right = False, labelleft = False )

        if savefig:

            plt.savefig( ip.output_directory + "FTIR/Plots/Plot.pdf" )

        else:

            plt.show()

        plt.close()

        if data_extraction_bool:

            array = data_extraction[0][:, np.newaxis]

            for i in range( 1, len( data_extraction ) ):

                array = np.hstack( (array, data_extraction[i][:, np.newaxis]) )

            np.savetxt( ip.output_directory + "Plot_Coords/Unnamed.txt", array )

    return shiny_de



# Imports.

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from .. import Global_Utilities as gu

# Function definitions.

def plot_data( ip, file_data, data, savefig = False ):

    resin_data = gu.get_list_of_resins_data( ip.directory )

    feature_names, features = gu.csv_to_df_to_array_and_column_titles( ip.output_directory + "Colour/Features/Features.csv" )

    sample, sample_array, samples_present, samples_present_array = gu.sample_data_from_file_data( file_data )

    samples_to_plot = samples_present
    samples_to_plot = [11, 14, 10, 4, 13, 21, 23, 18, 22, 20, 2, 3, 17, 16, 19, 1, 15, 12, 6, 5, 7, 9, 8, 24]
    samples_to_plot = [16, 17, 19, 24, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 18, 20, 21, 22, 23]

    specimens = False
    all_specimens = True
    specimen_mask = []

    mean = True

    if ip.shiny:

        samples_to_plot = ip.shiny_samples_to_plot

        if type( samples_to_plot ) == int:

            samples_to_plot = [samples_to_plot]

        specimen_mask = ip.shiny_specimens_to_plot

        if type( specimen_mask ) == int:

            specimen_mask = [specimen_mask]

        mean = ip.shiny_mean
        specimens = ip.shiny_specimen
        all_specimens = False

    if specimens:

        if ip.shiny:

            specimen_mask_2 = []

            for i in range( len( file_data ) ):

                if file_data[i][2] in specimen_mask:

                    specimen_mask_2.append( True )

                else:

                    specimen_mask_2.append( False )

            features = features[specimen_mask_2, :]

            return features[:, 0].tolist(), features[:, 1].tolist(), features[:, 2].tolist()

        specimen_mask = gu.produce_mask( sample_array, samples_to_plot )

        features = features[specimen_mask, :]
        file_data_mask = np.array( file_data )[specimen_mask] # Orders file data according to specimen mask.
        file_data_mask = [[int( f[0] ), int( f[1] ), f[2], f[3]] for f in file_data_mask] # Converts file data so that f[0], etc. are integers and not np.str.
        sample_array = sample_array[specimen_mask]

        ab_euclidean = np.sqrt( features[:, 1] * features[:, 1] + features[:, 2] * features[:, 2] )

        gu.plot_scatterplot_of_two_features( ip.directory, ab_euclidean, features[:, 0], sample_array, [f[2] for f in file_data], line_of_best_fit = False, xlabel = "sqrt(a*^2 + b*^2)", ylabel = "L*", savefig = savefig, filename = ip.output_directory + "Colour/Plots/L_vs_Colour.pdf" )

        gu.plot_scatterplot_of_three_features( ip.directory, features[:, 1], features[:, 2], features[:, 0], sample_array, [f[2] for f in file_data], title = "", xlabel = "a*", ylabel = "b*", zlabel = "L*", savefig = savefig, filename = ip.output_directory + "Colour/Plots/3D.pdf" )

    if mean:

        # resin_colours = pd.read_csv( ip.output_directory + "Colour/Sandbox/rgb_colours.csv", header = None )
        #
        # resin_colours[0] = resin_colours[0].astype( int )
        #
        # resin_colours.index = resin_colours[0]
        #
        # resin_colours.drop( columns = [0], inplace = True )
        #
        # resin_colours_list = [[resin_colours.loc[s, 1] / 255, resin_colours.loc[s, 2] / 255, resin_colours.loc[s, 3] / 255] for s in samples_to_plot]

        mean_features = gu.extract_mean_features( features, sample_array, samples_to_plot )

        if ip.shiny:

            return mean_features[:, 0], mean_features[:, 1], mean_features[:, 2]

        std_of_features = gu.extract_std_of_features( features, sample_array, samples_to_plot )

        ab_euclidean = np.sqrt( features[:, 1] * features[:, 1] + features[:, 2] * features[:, 2] )

        ab_euclidean_mean = gu.extract_mean_features( ab_euclidean[:, np.newaxis], sample_array, samples_to_plot )

        ab_euclidean_std = gu.extract_std_of_features( ab_euclidean[:, np.newaxis], sample_array, samples_to_plot )

        # gu.plot_scatterplot_of_two_features( ip.directory, ab_euclidean_mean[:, 0], mean_features[:, 0], samples_to_plot, [resin_data.loc[i]["Label"] for i in sample_mask], errorbars = True, std = [ab_euclidean_std[:, 0], std_of_features[:, 0]], line_of_best_fit = False, xlabel = "sqrt(a*^2 + b*^2)", ylabel = "L*", savefig = savefig, filename = output_directory + "Colour/Plots/L_vs_Colour.pdf" )

        # gu.plot_scatterplot_of_two_features( ip.directory, mean_features[:, 1], mean_features[:, 0], resin_colours_list, [resin_data.loc[i]["Label"] for i in samples_to_plot], line_of_best_fit = False, xlabel = "a*", ylabel = "L*", savefig = savefig, filename = ip.output_directory + "Colour/Plots/L_vs_a.pdf" )

        # gu.plot_scatterplot_of_two_features( ip.directory, mean_features[:, 2], mean_features[:, 0], resin_colours_list, [resin_data.loc[i]["Label"] for i in samples_to_plot], line_of_best_fit = False, xlabel = "b*", ylabel = "L*", savefig = savefig, filename = ip.output_directory + "Colour/Plots/L_vs_b.pdf" )

        gu.plot_scatterplot_of_three_features( ip.directory, mean_features[:, 1], mean_features[:, 2], mean_features[:, 0], samples_to_plot, [resin_data.loc[i]["Label"] for i in samples_to_plot], title = "", xlabel = "a*", ylabel = "b*", zlabel = "L*", savefig = savefig, filename = ip.output_directory + "Colour/Plots/3D.pdf" )

    return [], [], []


from setuptools import setup, find_packages
setup(
name='rHDPE_Data_Analysis',
version='1.0.110',
author='Phil Smith',
author_email='pvsmith100@btinternet.com',
description='Package for analysing data relating to recycled HDPE',
packages=find_packages(),
classifiers=[
'Programming Language :: Python :: 3',
'License :: OSI Approved :: MIT License',
'Operating System :: OS Independent',
],
python_requires='>=3.6',
)



# Imports.

import numpy as np
import pandas as pd

from . import DSC_plotting

from .. import Global_Utilities as gu

# Function definitions.

def compute_derivatives( data, width = 1 ):
    '''Compute the derivatives.'''

    derivative_data = [data[0][width: -width], [], data[2][width: -width], [], [], []]

    for i in range( len( data[1] ) ):

        derivative_data[1].append( np.array( gu.derivative( data[0], data[1][i], width ) ) )
        derivative_data[3].append( np.array( gu.derivative( data[2], data[3][i], width ) ) )

    for i in range( len( data[4] ) ):

        derivative_data[4].append( np.array( gu.derivative( data[0], data[4][i], width ) ) )
        derivative_data[5].append( np.array( gu.derivative( data[2], data[5][i], width ) ) )

    return derivative_data

def half_peak_width( temp, heat_flow, melt = False ):

    if not melt:

        peak = heat_flow.max()

    else:

        peak = heat_flow.min()

    half_peak = peak / 2

    if not melt:

        mask = np.where( heat_flow >= half_peak )[0]

    else:

        mask = np.where( heat_flow <= half_peak )[0]

    new_temp = temp[mask]

    temp_max = new_temp.max()
    temp_min = new_temp.min()

    return temp_max - temp_min

def extract_DSC_features( output_directory, file_data, data, first_derivative_data, second_derivative_data ):

    # Temp and Heat Flow of maximum of cryst second derivative between 120-130.

    feature_1, feature_2 = [], []

    temp_mask = np.where( (second_derivative_data[0] <= 130) & (second_derivative_data[0] >= 120) )[0]

    for i in range( len( file_data ) ):

        array = second_derivative_data[1][i][temp_mask]

        feature_1.append( array.max() )
        feature_2.append( second_derivative_data[0][temp_mask][np.argmax( array )] )

    features = np.array( feature_1 )[:, np.newaxis]

    features = np.hstack( (features, np.array( feature_2 )[:, np.newaxis]) )

    feature_names = ["DSC_HFC_120", "DSC_TC_120"]

    # Minimum of cryst first derivative between 80 and 90.

    feature_1 = []

    temp_mask = np.where( (first_derivative_data[0] <= 90) & (first_derivative_data[0] >= 80) )[0]

    for i in range( len( file_data ) ):

        array = first_derivative_data[1][i][temp_mask]

        feature_1.append( array.min() )

    features = np.hstack( (features, np.array( feature_1 )[:, np.newaxis]) )

    feature_names.append( "DSC_HFC_80" )

    # Temp and Heat Flow at maximum cryst heat flow and maximum melt heat flow.

    feature_1, feature_2, feature_3, feature_4 = [], [], [], []

    for i in range( len( file_data ) ):

        feature_1.append( data[0][np.argmax( data[1][i] )] )
        feature_2.append( data[1][i].max() )
        feature_3.append( data[2][np.argmin( data[3][i] )] )
        feature_4.append( data[3][i].min() )

    features = np.hstack( (features, np.array( feature_1 )[:, np.newaxis], np.array( feature_2 )[:, np.newaxis], np.array( feature_3 )[:, np.newaxis], np.array( feature_4 )[:, np.newaxis]) )

    feature_names.append( "DSC_TC_Max" )
    feature_names.append( "DSC_HFC_Max" )
    feature_names.append( "DSC_TM_Max" )
    feature_names.append( "DSC_HFM_Max" )

    # Max second derivative between 160 and 170.

    feature_1 = []

    temp_mask = np.where( (second_derivative_data[2] <= 170) & (second_derivative_data[2] >= 160) )[0]

    for i in range( len( file_data ) ):

        array = second_derivative_data[3][i][temp_mask]

        feature_1.append( array.max() )

    features = np.hstack( (features, np.array( feature_1 )[:, np.newaxis]) )

    feature_names.append( "DSC_HFM_160" )

    # Cryst and melt half-peak widths.

    feature_1, feature_2 = [], []

    for i in range( len( file_data ) ):

        feature_1.append( half_peak_width( data[0], data[1][i] ) )
        feature_2.append( half_peak_width( data[2], data[3][i], melt = True ) )

    features = np.hstack( (features, np.array( feature_1 )[:, np.newaxis], np.array( feature_2 )[:, np.newaxis]) )

    feature_names.append( "DSC_C_HalfPeak" )
    feature_names.append( "DSC_M_HalfPeak" )

    # Minimum of cryst second derivative between 102 and 106.

    feature_1 = []

    temp_mask = np.where( (second_derivative_data[0] <= 106) & (second_derivative_data[0] >= 102) )[0]

    for i in range( len( file_data ) ):

        array = second_derivative_data[1][i][temp_mask]

        feature_1.append( array.min() )

    features = np.hstack( (features, np.array( feature_1 )[:, np.newaxis]) )

    feature_names.append( "DSC_HFC_102" )

    # Cryst onset temperature.

    feature_1 = []

    for i in range( len( file_data ) ):

        feature_1.append( first_derivative_data[0][np.where( (first_derivative_data[1][i] < -0.2) )[0][0]] )

    features = np.hstack( (features, np.array( feature_1 )[:, np.newaxis]) )

    feature_names.append( "DSC_C_Onset" )

    # Melt onset temperature.

    feature_1 = []

    temp_mask = np.where( (first_derivative_data[2] <= 145) )[0]

    for i in range( len( file_data ) ):

        array = np.where( (first_derivative_data[3][i][temp_mask] > 0.1) )[0]

        feature_1.append( first_derivative_data[2][temp_mask][array[len( array ) - 1]] )

    features = np.hstack( (features, np.array( feature_1 )[:, np.newaxis]) )

    feature_names.append( "DSC_M_Onset" )

    # Crystallinity and PP content.

    feature_1, feature_2 = [], []

    temp_mask_1 = np.where( (data[2] <= 145) & (data[2] >= 65) )[0]
    temp_mask_2 = np.where( (data[2] <= 170) & (data[2] >= 143) )[0]

    for i in range( len( file_data ) ):

        integral_1 = gu.integral_2( data[2][temp_mask_1], -data[3][i][temp_mask_1], 0, len( data[2][temp_mask_1] ) - 1, min( -data[3][i][temp_mask_1][0], -data[3][i][temp_mask_1][len( data[3][i][temp_mask_1] ) - 1] ) ) / 293 * 6
        integral_2 = gu.integral_2( data[2][temp_mask_2], -data[3][i][temp_mask_2], 0, len( data[2][temp_mask_2] ) - 1, min( -data[3][i][temp_mask_2][0], -data[3][i][temp_mask_2][len( data[3][i][temp_mask_2] ) - 1] ) ) / 208 * 6 / 0.5

        if integral_2 < 0:

            integral_2 = 0

        # fpp = integral_2 / (integral_1 + integral_2)

        integral_1 = integral_1 / (1 - integral_2)

        integral_1 = integral_1 * 100
        integral_2 = integral_2 * 100

        feature_1.append( integral_1 )
        feature_2.append( integral_2 )

    features = np.hstack( (features, np.array( feature_1 )[:, np.newaxis], np.array( feature_2 )[:, np.newaxis]) )

    feature_names.append( "DSC_Crystallinity" )
    feature_names.append( "DSC_fPP" )

    # Modified melt temperature.

    feature_1 = []

    for i in range( len( file_data ) ):

        temp_mask_1 = np.where( (data[2] >= data[2][data[3][i].argmin()]) & (data[3][i] >= -2) )[0]

        feature_1.append( data[2][temp_mask_1[0]] )

    features = np.hstack( (features, np.array( feature_1 )[:, np.newaxis]) )

    feature_names.append( "DSC_Melt_Temp" )

    df = gu.array_with_column_titles_to_df( features, feature_names )

    df.to_csv( output_directory + "DSC/Features/Features.csv" )

# Crystallinity values from Andy.
#
# crystallinity = [0, 65.5, 64.3, 66, 64.1, 61.3, 61.8, 62.6, 58.5, 61.7, 60.7, 60, 68.2, 60.5, 65.2, 74.1, 66.3, 69, 59.4, 73, 62.4, 62.5, 61.2, 51.1, 66.7]

def read_and_analyse_features( ip, file_data ):

    plot_specimen_bars = False
    plot_mean_bars = False
    plot_specimen_features = True
    plot_mean_features = True
    plot_specimen_distance_matrix = True
    plot_mean_distance_matrix = True
    plot_specimen_dendrogram = True
    plot_mean_dendrogram = True

    resin_data = gu.get_list_of_resins_data( ip.directory )

    feature_names, features = gu.csv_to_df_to_array_and_column_titles( ip.output_directory + "DSC/Features/Features.csv" )

    sample, sample_array, samples_present, samples_present_array = gu.sample_data_from_file_data( file_data )

    if not ip.sample_mask:

        ip.sample_mask = [11, 14, 10, 4, 13, 21, 23, 18, 22, 20, 2, 3, 17, 16, 19, 1, 15, 12, 6, 5, 7, 9, 8, 24]
        # ip.sample_mask = [17, 16, 19, 24, 1, 12, 15, 6, 7, 2, 3, 5, 9, 8, 13, 11, 14, 20, 21, 10, 18, 4, 22, 23]
        # ip.sample_mask = [16, 17, 19, 24, 12, 15, 1, 2, 3, 5, 6, 7, 8, 9, 4, 10, 13, 11, 14, 18, 20, 21, 22, 23]

    sample_mask = ip.sample_mask

    sample_mask = gu.remove_redundant_samples( sample_mask, samples_present )

    if not ip.feature_selection:

        # ip.feature_selection = [i for i in range( len( feature_names ) )]
        ip.feature_selection = [7, 8, 9, 10, 13]
        # ip.feature_selection = [8, 9, 11, 12, 13, 14]
        # ip.feature_selection = [14, 13, 12, 11, 9, 8]

    feature_selection = ip.feature_selection

    features = features[:, feature_selection]
    feature_names = list( np.array( feature_names )[feature_selection] )

    mean_features_unnormalised = gu.extract_mean_features( features, sample_array, samples_present )
    mean_feature_names = feature_names.copy()

    mean_features_unnormalised_plus_sample_mask = np.hstack( (samples_present_array[:, np.newaxis], mean_features_unnormalised) )

    df = gu.array_with_column_titles_to_df( mean_features_unnormalised_plus_sample_mask, ["sample"] + mean_feature_names )

    df.to_csv( ip.output_directory + "DSC/Features/Mean_Features_Unnormalised.csv" )

    std_of_features = gu.extract_std_of_features( features, sample_array, samples_present )

    std_of_features_plus_sample_mask = np.hstack( (samples_present_array[:, np.newaxis], std_of_features) )

    df = gu.array_with_column_titles_to_df( std_of_features_plus_sample_mask, ["sample"] + mean_feature_names )

    df.to_csv( ip.output_directory + "DSC/Features/Std_of_Features_Unnormalised.csv" )

    specimen_mask = gu.produce_mask( sample_array, sample_mask )

    features = features[specimen_mask, :]
    file_data_mask = np.array( file_data )[specimen_mask] # Orders file data according to specimen mask.
    file_data_mask = [[int( f[0] ), int( f[1] ), f[2], f[3]] for f in file_data_mask] # Converts file data so that f[0], etc. are integers and not np.str.
    sample_array = sample_array[specimen_mask]

    feature_weights = [1 for i in range( len( feature_selection ) )]

    gu.normalise_features( features, feature_weights )

    mean_features = gu.extract_mean_features( features, sample_array, sample_mask )
    mean_feature_names = feature_names.copy()

    std_of_features = gu.extract_std_of_features( features, sample_array, sample_mask )

    distance_matrix = gu.distance_matrix_from_features( features )

    mean_distance_matrix = gu.distance_matrix_from_features( mean_features )

    mean_features_plus_sample_mask = np.hstack( (np.array( sample_mask )[:, np.newaxis], mean_features) )

    df = gu.array_with_column_titles_to_df( mean_features_plus_sample_mask, ["sample"] + mean_feature_names )

    df.to_csv( ip.output_directory + "DSC/Features/Mean_Features.csv" )

    std_of_features_plus_sample_mask = np.hstack( (np.array( sample_mask )[:, np.newaxis], std_of_features) )

    df = gu.array_with_column_titles_to_df( std_of_features_plus_sample_mask, ["sample"] + mean_feature_names )

    df.to_csv( ip.output_directory + "DSC/Features/Std_of_Features.csv" )

    df = gu.array_with_column_titles_and_label_titles_to_df( mean_distance_matrix, sample_mask, sample_mask )

    df.to_csv( ip.output_directory + "DSC/Features/Distance_Matrix.csv" )

    # Perform PCA for DSC features.

    # gu.perform_pca( ip.directory, gu.array_with_column_titles_to_df( mean_features, [str( i ) for i in mean_feature_names] ), sample_mask, std_error = True, std_of_features_df = pd.DataFrame( std_of_features ), num_components = 2, filename = ip.output_directory + "DSC/Features/PCA.pdf" )

    if plot_specimen_bars:

        for i in range( len( features[0] ) ):

            gu.plot_barchart_of_feature( features[:, i], [f[2] for f in file_data_mask], colour = True, colour_mask = sample_array, filename = ip.output_directory + "DSC/Feature_Bars/Specimen/" + feature_names[i] + ".pdf", savefig = True )

    if plot_mean_bars:

        for i in range( len( mean_features[0] ) ):

            gu.plot_barchart_of_feature( mean_features[:, i], [resin_data.loc[i]["Label"] for i in sample_mask], errorbars = True, std = std_of_features[:, i], colour = True, colour_mask = sample_mask, filename = ip.output_directory + "DSC/Feature_Bars/Mean/" + mean_feature_names[i] + ".pdf", savefig = True )

    if plot_specimen_features:

        gu.plot_features( ip.output_directory, features, feature_names, [f[2] for f in file_data_mask], specimen = True, subdirectory = "DSC/Features/", title = "Specimen_Features.pdf" )

    if plot_mean_features:

        gu.plot_features( ip.output_directory, mean_features, mean_feature_names, [resin_data.loc[i]["Label"] for i in sample_mask], subdirectory = "DSC/Features/", title = "Means_Features.pdf" )

    if plot_specimen_distance_matrix:

        gu.plot_distance_matrix( ip.output_directory, distance_matrix, [f[2] for f in file_data_mask], specimen = True, file_data = file_data_mask, sample_mask = sample_mask, subdirectory = "DSC/Features/", title = "Specimen_Distance_Matrix.pdf" )

    if plot_mean_distance_matrix:

        gu.plot_distance_matrix( ip.output_directory, mean_distance_matrix, [resin_data.loc[i]["Label"] for i in sample_mask], subdirectory = "DSC/Features/", title = "Means_Distance_Matrix.pdf" )

    if plot_specimen_dendrogram:

        gu.plot_dendrogram( ip.output_directory, distance_matrix, [f[2] for f in file_data_mask], specimen = True, subdirectory = "DSC/Features/", title = "Specimen_Dendrogram.pdf" )

    if plot_mean_dendrogram:

        gu.plot_dendrogram( ip.output_directory, mean_distance_matrix, [resin_data.loc[i]["Label"] for i in sample_mask], subdirectory = "DSC/Features/", title = "Means_Dendrogram.pdf" )

def variance_analysis( directory, output_directory, file_data, data ):

    resin_data = gu.get_list_of_resins_data( directory )

    sample, sample_array, samples_present, samples_present_array = gu.sample_data_from_file_data( file_data )

    sample_mask = [11, 14, 10, 4, 13, 21, 23, 18, 22, 20, 2, 3, 17, 16, 19, 1, 15, 12, 6, 5, 7, 9, 8, 24]

    sample_mask = gu.remove_redundant_samples( sample_mask, samples_present )

    integral_c, integral_m = np.zeros( len( sample_mask ) ), np.zeros( len( sample_mask ) )

    for i, s in enumerate( sample_mask ):

        mask = np.where( sample_array == s )[0]

        range_c, high_point_c, low_point_c, median_c, mean_c, std_c = gu.compute_range_mean_std( data[1], mask )
        range_m, high_point_m, low_point_m, median_c, mean_m, std_m = gu.compute_range_mean_std( data[3], mask )

        DSC_plotting.plot_variance( directory, output_directory + "DSC/Sandbox/Variance/", data, mask, s, std_c, std_m )

        if len( mask ) >= 3:

            integral_c[i] = gu.integral_3( np.array( std_c ), 0.1 )
            integral_m[i] = gu.integral_3( np.array( std_m ), 0.1 )

    nonzero_mask = np.nonzero( integral_c )[0]
    integral_c = integral_c[nonzero_mask]
    integral_m = integral_m[nonzero_mask]
    sample_mask = np.array( sample_mask )[nonzero_mask]

    gu.plot_barchart_of_feature( integral_c, [resin_data.loc[i]["Label"] for i in sample_mask], colour = True, colour_mask = sample_mask, xlabel = "PCR Sample", ylabel = "Integral of Standard Deviation", title = "Variance of Samples during Crystallisation", filename = output_directory + "DSC/Sandbox/Variance/BarCryst.pdf", savefig = True )

    gu.plot_barchart_of_feature( integral_m, [resin_data.loc[i]["Label"] for i in sample_mask], colour = True, colour_mask = sample_mask, xlabel = "PCR Sample", ylabel = "Integral of Standard Deviation", title = "Variance of Samples during Melt", filename = output_directory + "DSC/Sandbox/Variance/BarMelt.pdf", savefig = True )

def identify_anomalies( file_data, data ):

    sample, sample_array, samples_present, samples_present_array = gu.sample_data_from_file_data( file_data )

    sample_mask = [11, 14, 10, 4, 13, 21, 23, 18, 22, 20, 2, 3, 17, 16, 19, 1, 15, 12, 6, 5, 7, 9, 8, 24]

    sample_mask = gu.remove_redundant_samples( sample_mask, samples_present )

    for s in sample_mask:

        mask = np.where( sample_array == s )[0]

        if len( mask ) < 3:

            print( "Less than three specimens of sample {} so no specimens removed".format( s ) )

            continue

        for i in mask:

            anomalous_cryst, anomalous_melt = False, False

            other_specimens = [j for j in mask if j != i]

            range_s, high_point, low_point, median, mean, std = gu.compute_range_mean_std( data[1], other_specimens )

            for j in range( len( data[1][0] ) ):

                value = data[1][i][j]

                error = 0.6 + 3 * std[j]

                if value < mean[j] - error or value > mean[j] + error:

                    anomalous_cryst = True

                    print( file_data[i][2], "should be removed due to value {} during crystallisation at {}".format( value, data[0][j] ) )

                    for l in mask:

                        print( data[1][l][j] )

                    print( mean[j], error )

                    break

            range_s, high_point, low_point, median, mean, std = gu.compute_range_mean_std( data[3], other_specimens )

            for j in range( len( data[3][0] ) ):

                value = data[3][i][j]

                error = 0.6 + 3 * std[j]

                if value < mean[j] - error or value > mean[j] + error:

                    anomalous_melt = True

                    print( file_data[i][2], "should be removed due to value {} during melt at {}".format( value, data[2][j] ) )

                    for l in mask:

                        print( data[3][l][j] )

                    print( mean[j], error )

                    break

            if anomalous_melt and anomalous_cryst:

                print( file_data[i], "should be removed." )

def boxplot_variance( directory, output_directory, file_data, data ):

    sample, sample_array, samples_present, samples_present_array = gu.sample_data_from_file_data( file_data )

    sample_mask = [11, 14, 10, 4, 13, 21, 23, 18, 22, 20, 2, 3, 17, 16, 19, 1, 15, 12, 6, 5, 7, 9, 8, 24]

    feature_names, features = gu.csv_to_df_to_array_and_column_titles( output_directory + "DSC/Features/Features.csv" )
    # _, std_of_features = gu.csv_to_df_to_array_and_column_titles( output_directory + "DSC/Features/Std_of_Features.csv" )

    feature_selection = [8, 9, 11, 12, 13, 14]

    features = features[:, feature_selection]
    feature_names = list( np.array( feature_names )[feature_selection] )

    range_of_features = gu.extract_range_of_features( features, sample_array, sample_mask )

    print( feature_names )
    print( np.ptp(features, axis = 0))
    print( np.mean( range_of_features, axis = 0))
    print(np.mean( range_of_features, axis = 0) / np.ptp(features, axis = 0))

    sample_mask = [11, 14, 10, 4, 13, 21, 23, 18, 22, 20, 2, 3, 17, 16, 19, 1, 15, 12, 6, 5, 7, 9, 8, 24]

    gu.plot_boxplot( directory, features, feature_names, file_data, sample_mask, "DSC_C_Onset", xlabel = "Samples", ylabel = "Secondary Peak at 160°C" )

def comparison_of_different_machines( directory, output_directory, file_data ):

    resin_data = gu.get_list_of_resins_data( directory )

    feature_names, features = gu.csv_to_df_to_array_and_column_titles( output_directory + "DSC/Features/Features.csv" )

    sample, sample_array, samples_present, samples_present_array = gu.sample_data_from_file_data( file_data )

    feature_weights = [1 for i in range( len( features[0] ) )]

    # gu.normalise_features( features, feature_weights )

    mean_features = gu.extract_mean_features( features, sample_array, samples_present )
    mean_feature_names = feature_names.copy()

    std_of_features = gu.extract_std_of_features( features, sample_array, samples_present )

    samples_to_compare = [1, 2, 3, 4, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24]
    feature_name = "DSC_M_Onset"

    x_values = [mean_features[samples_present.index( i )][feature_names.index( feature_name )] for i in samples_to_compare]
    y_values = [mean_features[samples_present.index( i + 100 )][feature_names.index( feature_name )] for i in samples_to_compare]

    gu.plot_scatterplot_of_two_features( x_values, y_values, samples_to_compare, [resin_data.loc[i]["Label"] for i in samples_to_compare], line_of_best_fit = False, xlabel = "PP Content from Machine 1", ylabel = "PP Content from Machine 2" )

def sandbox( directory, output_directory, file_data, data, first_derivative_data, second_derivative_data ):

    perform_variance_analysis = False # Perform computations relating to variance.

    perform_identify_anomalies = False # Identifies anomalies based on standard deviations.

    perform_boxplot_variance = True # Plot boxplots of variance of features for specimens.

    perform_comparison_of_different_machines = False # Compare curves obtained from different machines.

    if perform_variance_analysis:

        variance_analysis( directory, output_directory, file_data, data )

    if perform_identify_anomalies:

        identify_anomalies( file_data, data )

    if perform_boxplot_variance:

        boxplot_variance( directory, output_directory, file_data, data )

    if perform_comparison_of_different_machines:

        comparison_of_different_machines( directory, output_directory, file_data )



# Imports.

import numpy as np
import matplotlib.pyplot as plt

from .. import Global_Utilities as gu

# Function definitions.

def plot_data( ip, file_data, data, first_derivative_data, second_derivative_data, third_derivative_data, savefig = False ):

    sample, sample_array, samples_present, samples_present_array = gu.sample_data_from_file_data( file_data )

    samples_to_plot = [18]

    specimens = True
    all_specimens = False
    specimen_mask_by_index = [6, 7, 8, 9, 10, 11]
    specimen_mask = []

    mean = False

    deriv0 = True
    deriv1 = False
    deriv2 = False
    deriv3 = False

    split = True
    # splits = [1.5, 2, 2.5, 3, 3.5, 4, 4.5, 5, 5.5, 6, 6.5, 7, 7.5, 8, 8.5, 9, 9.5, 10, 15]
    splits = [4, 4.3]

    if not split:

        splits = [int( data[0][0] ), int( data[0][len( data[0] ) - 1] )]

    if ip.shiny:

        samples_to_plot = ip.shiny_samples_to_plot

        if type( samples_to_plot ) == int:

            samples_to_plot = [samples_to_plot]

        specimen_mask = ip.shiny_specimens_to_plot

        if type( specimen_mask ) == int:

            specimen_mask = [specimen_mask]

        splits = ip.shiny_split
        mean = ip.shiny_mean
        specimens = ip.shiny_specimen
        all_specimens = False

    colours = gu.read_list_of_colours( ip.directory )

    shiny_de = []

    data_extraction_bool = False

    for s in range( len( splits ) - 1 ):

        data_extraction = []

        lower_bound, upper_bound = splits[s], splits[s + 1]

        for i in samples_to_plot:

            if specimens:

                mask = np.where( sample_array == i )[0]

                if not ip.shiny:

                    if all_specimens:

                        specimen_mask = [file_data[mask[j]][2] for j in range( len( mask ) )]

                    else:

                        specimen_mask = [file_data[mask[j]][2] for j in specimen_mask_by_index]

                for j in mask:

                    if file_data[j][2] in specimen_mask:

                        if deriv0:

                            rt_mask = np.where( (data[0] <= upper_bound) & (data[0] >= lower_bound) )[0]

                            plt.plot( data[0][rt_mask], data[1][j][rt_mask] / 10000000, label = file_data[j][2] )

                            shiny_de.append( data[0][rt_mask].tolist() )
                            shiny_de.append( (data[1][j][rt_mask]).tolist() )

            if mean:

                index = np.where( samples_present_array == i )[0][0]

                if deriv0:

                    rt_mask = np.where( (data[0] <= upper_bound) & (data[0] >= lower_bound) )[0]

                    plt.plot( data[0][rt_mask], data[2][index][rt_mask], label = i, color = colours[i] )

                    shiny_de.append( data[0][rt_mask].tolist() )
                    shiny_de.append( data[2][index][rt_mask].tolist() )

        if ip.shiny:

            return shiny_de

        plt.legend( ncol = 1, bbox_to_anchor = ( 1.05, 1 ), loc = 'upper left', borderaxespad = 0 )
        # plt.legend()

        plt.xlabel( "Retention Time [Minutes]" )
        plt.ylabel( "Absolute Intensity [x10e7]" )

        ax = plt.gca()

        plt.tight_layout()

        if savefig:

            plt.savefig( ip.output_directory + "Plots/Plot.pdf" )

        else:

            plt.show()

        plt.close()

        if data_extraction_bool:

            array = data_extraction[0][:, np.newaxis]

            for i in range( 1, len( data_extraction ) ):

                array = np.hstack( (array, data_extraction[i][:, np.newaxis]) )

            np.savetxt( ip.output_directory + "Output/Plot_Coords/Unnamed.txt", array )

    return shiny_de



# Imports.

import os
import glob
import re
import numpy as np
import pandas as pd

from .. import Global_Utilities as gu

# Function definitions.

def read_raw_data_file_1( filename, f, resin_data, file_data, data ):

    pattern = re.compile( r"^Resin(\d+)_(\d+)_" )

    resin = int( pattern.search( f ).groups()[0] )

    specimen = int( pattern.search( f ).groups()[1] )

    with open( filename, 'r', encoding = 'utf-16' ) as file:

        column_data = [[], [], [], [], []]

        lines = file.readlines()

        linenumber = 0

        for line in lines:

            if linenumber < 10:

                linenumber += 1
                continue

            a_list = line.split()

            column_data[0].append( float( a_list[1] ) )
            column_data[1].append( float( a_list[3] ) )
            column_data[2].append( float( a_list[4] ) )
            column_data[3].append( float( a_list[2] ) )
            column_data[4].append( float( a_list[5] ) )

    for i in range( 5 ):

        data[i].append( np.array( column_data[i] ) )

    file_data.append( [resin, specimen, resin_data.loc[resin]["Label"] + ".{}".format( specimen ), ""] )

def read_raw_data_file_2( filename, f, resin_data, file_data, data ):

    pattern = re.compile( r"^Resin(\d+)_(\d+)_" )

    resin = int( pattern.search( f ).groups()[0] )

    specimen = int( pattern.search( f ).groups()[1] )

    with open( filename, 'r', encoding = "windows-1252" ) as file:

        column_data = [[], [], [], [], []]

        lines = file.readlines()

        linenumber = 0

        for line in lines:

            if linenumber < 2:

                linenumber += 1
                continue

            a_list = line.rstrip().split( "," )

            column_data[0].append( float( a_list[3] ) )
            column_data[1].append( float( a_list[0] ) * 10 ** 6 )
            column_data[2].append( float( a_list[1] ) * 10 ** 6 )
            column_data[3].append( float( a_list[9] ) )
            column_data[4].append( float( a_list[2] ) )

    for i in range( 5 ):

        column_data[i].reverse()
        data[i].append( np.array( column_data[i] ) )

    file_data.append( [resin, specimen, resin_data.loc[resin]["Label"] + ".{}".format( specimen ), ""] )

def extract_raw_data( directory, data_directory ):
    '''Extract the raw data from the files.'''

    resin_data = gu.get_list_of_resins_data( directory ) # Obtain the spreadsheet of data for the resins.

    resins = sorted( [os.path.basename( path ) for path in glob.glob( data_directory + "*" )], key = gu.sort_raw_files_1 )

    file_data, data = [], [[], [], [], [], []]

    pattern = re.compile( r"^Resin(\d+)" )

    for r in resins:

        filenames = sorted( [os.path.basename( path ) for path in glob.glob( data_directory + r + "/*" )], key = gu.sort_raw_files_2 )

        resin = int( pattern.search( r ).groups()[0] )

        for f in filenames:

            if resin == 40 or resin == 41:

                read_raw_data_file_2( data_directory + r + "/" + f, f, resin_data, file_data, data )

            else:

                read_raw_data_file_1( data_directory + r + "/" + f, f, resin_data, file_data, data )

    return file_data, data

def standardise_data( data ):
    '''Standardise data.'''

    standardised_angular_frequency = [10 ** (i / 10) for i in range( -10, 21 )]
    standardised_angular_frequency.reverse()

    for i in range( len ( data[0] ) ):

        sample_data = [[], [], [], [], []]

        for j in standardised_angular_frequency:

            index_1, index_2 = -2, -1

            for k, v in enumerate( data[0][i] ):

                if v < j:

                    index_1 = k - 1
                    index_2 = k
                    break

            log_change = (np.log10( j ) - np.log10( data[0][i][index_1] )) / (np.log10( data[0][i][index_2] ) - np.log10( data[0][i][index_1] ))

            for l in range( 1, 5 ):

                if data[l][i].any():

                    sample_data[l].append( 10 ** ((np.log10( data[l][i][index_2] ) - np.log10( data[l][i][index_1] )) * log_change + np.log10( data[l][i][index_1] )) )

        for l in range( 1, 5 ):

            data[l][i] = np.array( sample_data[l] )

    data[0] = np.array( standardised_angular_frequency )

def add_additional_variables( data ):

    complex_viscosity, loss_factor = [], []

    for i in range( len( data[1] ) ):

        cv, lf = [], []

        for j in range( len( data[1][i] ) ):

            cv.append( np.sqrt( data[1][i][j] ** 2 * data[2][i][j] ** 2 ) / data[0][j] )
            lf.append( data[2][i][j] / data[1][i][j] )

        complex_viscosity.append( cv )
        loss_factor.append( lf )

    data[3] = np.array( complex_viscosity )
    data[4] = np.array( loss_factor )

def add_description_to_file_data( file_data ):
    '''Add descriptions in the form of letters to each specimen.'''

    pass

def read_files_and_preprocess( directory, data_directory, merge_groups ):
    '''Read files and preprocess data.'''

    file_data, data = extract_raw_data( directory, data_directory )

    standardise_data( data )

    # add_additional_variables( data ) # Uncomment to compute additional variables from storage and loss modulii directly.

    add_description_to_file_data( file_data )

    if merge_groups:

        gu.merge( file_data )

    return file_data, data

def write_csv( output_directory, file_data, data, name_appendage = "" ):
    '''Write read and preprocessed data to a .csv file.'''

    array = data[0][:, np.newaxis]

    for i in range( len( data[1] ) ):

        array = np.hstack( (array, data[1][i][:, np.newaxis]) )
        array = np.hstack( (array, data[2][i][:, np.newaxis]) )
        array = np.hstack( (array, data[3][i][:, np.newaxis]) )
        array = np.hstack( (array, data[4][i][:, np.newaxis]) )

    np.savetxt( output_directory + "Rheology/Condensed_Data/Rheology_data" + name_appendage + ".csv", array, delimiter = ",", fmt = "%.5f" )

    array = np.array( file_data )

    np.savetxt( output_directory + "Rheology/Condensed_Data/file_data" + name_appendage + ".csv", array, delimiter = ",", fmt = "%s" )

def read_csv( directory, output_directory, merge_groups, name_appendage = "" ):
    '''Read the preprocessed .csv files.'''

    resin_data = gu.get_list_of_resins_data( directory ) # Obtain the spreadsheet of data for the resins.

    file_data = []

    df = pd.read_csv( output_directory + "Rheology/Condensed_Data/file_data" + name_appendage + ".csv", sep = ",", header = None )

    for i in range( len( df.index ) ):

        resin = df.iloc[i, 0]
        specimen = df.iloc[i, 1]

        file_data.append( [resin, specimen, resin_data.loc[resin]["Label"] + ".{}".format( specimen ), ""] )

    data = [[], [], [], [], []]

    df = pd.read_csv( output_directory + "Rheology/Condensed_Data/Rheology_data" + name_appendage + ".csv", sep = ",", header = None )

    data[0] = df.iloc[:, 0].to_numpy( dtype = np.float64 )

    for i in range( 1, len( df.columns ) ):

        if i % 4 == 1:

            data[1].append( df.iloc[:, i].to_numpy( dtype = np.float64 ) )

        elif i % 4 == 2:

            data[2].append( df.iloc[:, i].to_numpy( dtype = np.float64 ) )

        elif i % 4 == 3:

            data[3].append( df.iloc[:, i].to_numpy( dtype = np.float64 ) )

        elif i % 4 == 0:

            data[4].append( df.iloc[:, i].to_numpy( dtype = np.float64 ) )

    add_description_to_file_data( file_data )

    if merge_groups:

        gu.merge( file_data )

    return file_data, data

def remove_files( file_data, data, descriptors_to_remove = "" ):
    '''Remove files not needed/wanted for analysis by searching for letters in file descriptions.'''

    files_to_remove = []

    for i in range( len( file_data ) ):

        s = file_data[i][3]

        for l in descriptors_to_remove:

            if s.find( l ) > -0.5:

                files_to_remove.append( i )

                break

    files_to_remove.reverse()

    for r in files_to_remove:

        file_data.pop( r )
        data[1].pop( r )
        data[2].pop( r )
        data[3].pop( r )
        data[4].pop( r )

    return file_data, data

def compute_mean( output_directory, file_data, data, name_appendage = "" ):
    '''Compute the mean data for each resin.'''

    m = [gu.sample_mean( file_data, data[1] ), gu.sample_mean( file_data, data[2] ), gu.sample_mean( file_data, data[3] ), gu.sample_mean( file_data, data[4] )]

    labels = ["SM", "LM", "CV", "LF"]

    for i in range( len( m ) ):

        array = m[i][0][:, np.newaxis]

        for j in range( 1, len( m[i] ) ):

            array = np.hstack( (array, m[i][j][:, np.newaxis]) )

        np.savetxt( output_directory + "Rheology/Condensed_Data/" + labels[i] + "_Means" + name_appendage + ".csv", array, delimiter = ",", fmt = "%.5f" )

def read_mean( output_directory, data, name_appendage = "" ):
    '''Read the computed means for each resin from a file.'''

    labels = ["SM", "LM", "CV", "LF"]

    for i in range( len( labels ) ):

        m = []

        df = pd.read_csv( output_directory + "Rheology/Condensed_Data/" + labels[i] + "_Means" + name_appendage + ".csv", sep = ",", header = None )

        for j in range( len( df.columns ) ):

            m.append( df.iloc[:, j].to_numpy( dtype = np.float64 ) )

        if len( data ) > 5 + i: # A previous list of means is already appended.

            data[5 + i] = m

        else:

            data.append( m )

    return data



# Imports.

from numbers_parser import Document

# Class definitions.

class Input_Parameters():

    def __init__( self ):

        self.shiny = False

        self.directory = ""

        self.output_directory = ""

        self.datasets_to_read = []

        self.sample_mask = []

        self.read_files = False

        self.plot_global_features = False

        self.scatterplot = False

        self.correlation_heatmaps = False

        self.pca = False

        self.distance_to_virgin_analysis_based_on_pcas = False

        self.rank_resins_by_pp_content = False

        self.manual_ml = False

        self.pca_ml = False

        self.sandbox = False

        self.scatterplot_x = ""

        self.scatterplot_y = ""

# Function definitions.

def read_parameters_from_numbers_file( filename, ip ):

    doc = Document( filename )

    sheets = doc.sheets
    tables = sheets[0].tables
    rows = tables[0].rows()

    ip.shiny = bool( rows[0][1].value )

    ip.directory = str( rows[1][1].value or "" )

    ip.output_directory = str( rows[2][1].value or "" )

    ip.datasets_to_read = [int( i.value ) for i in rows[3][1:] if i.value != None]

    ip.sample_mask = [int( i.value ) for i in rows[4][1:] if i.value != None]

    ip.read_files = bool( rows[5][1].value )

    ip.plot_global_features = bool( rows[6][1].value )

    ip.scatterplot = bool( rows[7][1].value )

    ip.correlation_heatmaps = bool( rows[8][1].value )

    ip.pca = bool( rows[9][1].value )

    ip.distance_to_virgin_analysis_based_on_pcas = bool( rows[10][1].value )

    ip.rank_resins_by_pp_content = bool( rows[11][1].value )

    ip.manual_ml = bool( rows[12][1].value )

    ip.pca_ml = bool( rows[13][1].value )

    ip.sandbox = bool( rows[14][1].value )
